/*
 * GImageView
 * Copyright (C) 2001 Takuro Ashie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#include "gimageview.h"

#include "gfileutil.h"
#include "gimv_image.h"
#include "fileutil.h"
#include "image_info.h"
#include "prefs.h"
#include "thumbnail_support.h"


GHashTable *ImageInfoTable = NULL;


/******************************************************************************
 *
 *   Private Functions.
 *
 ******************************************************************************/
static ImageInfo *
image_info_new (const gchar *filename)
{
   ImageInfo *info;

   info = g_new0 (ImageInfo, 1);
   g_return_val_if_fail (info, NULL);

   info->filename     = g_strdup (filename);
   info->archive_name = NULL;
   info->archive      = NULL;
   info->link         = NULL;
   info->format       = gimv_image_detect_type_by_ext (filename);
   info->width        = -1;
   info->height       = -1;
   info->depth        = -1;
   info->comment      = NULL;
   info->synced_to_orig_image = FALSE;

   info->ref_count = 1;

   return info;
}


/******************************************************************************
 *
 *   Public Functions.
 *
 ******************************************************************************/
/*
 *  image_info_get:
 *     @ Get infomation about normal file from cache.
 *       If it doesn't exist, create it.
 *
 *  filename:
 *  Return:
 */
ImageInfo *
image_info_get (const gchar *filename)
{
   ImageInfo *info;
   struct stat st;

   g_return_val_if_fail (filename, NULL);

   if (!ImageInfoTable) {
      ImageInfoTable = g_hash_table_new (g_str_hash, g_str_equal);
   }

   info = g_hash_table_lookup (ImageInfoTable, filename);
   if (!info) {
      if (!file_exists (filename)) return NULL;
      info = image_info_new (filename);
      if (!info) return NULL;
      stat (filename, &info->st);
      g_hash_table_insert (ImageInfoTable, info->filename, info);
   } else {
      if (stat (filename, &st)) return NULL;
      if (info->st.st_size != st.st_size
	  || info->st.st_mtime != st.st_mtime
	  || info->st.st_ctime != st.st_ctime)
      {
	 memcpy (&info->st, &st, sizeof (struct stat));
	 info->width = -1;
	 info->height = -1;
	 info->synced_to_orig_image = FALSE;
      }
      image_info_ref (info);
   }

   return info;
}


ImageInfo *
image_info_get_with_archive (const gchar *filename,
			     FRArchive   *archive,
			     struct stat *st)
{
   ImageInfo *info;
   gchar buf[MAX_PATH_LEN];

   g_return_val_if_fail (filename, NULL);
   g_return_val_if_fail (archive, NULL);

   if (!ImageInfoTable) {
      ImageInfoTable = g_hash_table_new (g_str_hash, g_str_equal);
   }

   g_snprintf (buf, MAX_PATH_LEN, "%s/%s", archive->filename, filename);

   info = g_hash_table_lookup (ImageInfoTable, buf);
   if (!info) {
      info = image_info_new (filename);
      info->archive_name = g_strdup (archive->filename);
      info->archive = archive;
      g_hash_table_insert (ImageInfoTable, g_strdup (buf), info);
   } else {
      /* FIXME!! update if necessary */
      /* image_info_ref (info); */
   }

   if (st)
      memcpy (&info->st, st, sizeof (struct stat));

   return info;
}


ImageInfo *
image_info_lookup (const gchar *filename)
{
   ImageInfo *info;

   info = g_hash_table_lookup (ImageInfoTable, filename);

   if (info)
      image_info_ref (info);

   return info;
}


void
image_info_finalize (ImageInfo *info)
{
   guint num;
   gchar buf[MAX_PATH_LEN];

   g_return_if_fail (info);

   /* remove data from hash table */
   if (info->archive) {
      gchar *orig_key;
      ImageInfo *value;
      gboolean success;

      g_snprintf (buf, MAX_PATH_LEN, "%s/%s",
		  info->archive->filename, info->filename);
      success = g_hash_table_lookup_extended (ImageInfoTable, buf,
					      (gpointer) &orig_key,
					      (gpointer) &value);
      if (success) {
	 g_hash_table_remove (ImageInfoTable, buf);
	 g_free (orig_key);
      }
   } else {
      g_hash_table_remove (ImageInfoTable, info->filename);
   }

   if (info->filename) {
      g_free (info->filename);
      info->filename = NULL;
   }
   if (info->archive_name) {
      g_free (info->archive_name);
      info->archive_name = NULL;
   }
   if (info->archive) {
      info->archive = NULL;
   }
   g_free (info);

   /* destroy hash table if needed */
   num = g_hash_table_size (ImageInfoTable);
   if (num < 1) {
      g_hash_table_destroy (ImageInfoTable);
      ImageInfoTable = NULL;
   }
}


ImageInfo *
image_info_ref (ImageInfo *info)
{
   g_return_val_if_fail (info, NULL);

   info->ref_count++;

   if (info->archive)
      gtk_object_ref (GTK_OBJECT (info->archive));

   return info;
}


void
image_info_unref (ImageInfo *info)
{
   FRArchive *archive;

   g_return_if_fail (info);

   archive = info->archive;

   info->ref_count--;

   if (info->ref_count < 1) {
      image_info_finalize (info);
   }

   if (archive)
      gtk_object_unref (GTK_OBJECT (archive));
}


/* used by fr-command */
void
image_info_unref_with_archive (ImageInfo *info)
{
   g_return_if_fail (info);

   info->ref_count--;
   /* info->archive = NULL; */

   if (info->ref_count < 1)
      image_info_finalize (info);
}


gboolean
image_info_set_data_from_image (ImageInfo *info, GimvImage *image)
{
   gboolean is_exist;

   g_return_val_if_fail (info, FALSE);

   if (!image) return FALSE;

   is_exist = !stat (info->filename, &info->st);
   if (!is_exist) return FALSE;

   gimv_image_get_size (image, &info->width, &info->height);
   info->synced_to_orig_image = TRUE;

   return TRUE;
}


gboolean
image_info_is_in_archive (ImageInfo *info)
{
   g_return_val_if_fail (info, FALSE);

   if (info->archive)
      return TRUE;
   else
      return FALSE;
}


const gchar *
image_info_get_path (ImageInfo *info)
{
   g_return_val_if_fail (info, NULL);

   return info->filename;
}


gchar *
image_info_get_path_with_archive (ImageInfo *info)
{
   g_return_val_if_fail (info, NULL);

   if (!info->archive) return NULL;

   return g_strconcat (info->archive_name, "/", info->filename, NULL);
}


const gchar *
image_info_get_archive_path (ImageInfo *info)
{
   g_return_val_if_fail (info, NULL);
   g_return_val_if_fail (info->archive, NULL);

   return info->archive->filename;
}


/*
 *  image_info_extract_archive:
 *     @ extract an image from archive.
 *
 *  thumb   :
 *  archive :
 *  Return  : Temporary file name. (shoud be free)
 */
/*
 *  FIXME!!
 *  If archive was already destroyed, create FRArchive object first.
 *  To detect whether archive was destroed or not, check info->archive.
 *  When archive is destroyed, fr-archive class will set this value to NULL
 *  by using image_info_unref_with_archive ().
 */
gchar *
image_info_extract_archive (ImageInfo *info)
{
   FRArchive *archive;
   GList *filelist = NULL;
   gchar *temp_dir;
   gchar *filename, *temp_file, buf[MAX_PATH_LEN];

   g_return_val_if_fail (info, NULL);

   archive = info->archive;
   g_return_val_if_fail (archive, NULL);

   filename = info->filename;

   temp_dir = gtk_object_get_data (GTK_OBJECT (archive), "temp-dir");

   g_return_val_if_fail (temp_dir && *temp_dir, NULL);

   g_snprintf (buf, MAX_PATH_LEN, "%s/%s",
	       temp_dir, filename);
   temp_file = buf;

   if (!file_exists (temp_file) && !archive->process->running) {
      ensure_dir_exists (temp_dir);
      filelist = g_list_append (filelist, filename);

      fr_archive_extract (archive, filelist, temp_dir,
			  FALSE, TRUE, FALSE);

      gtk_main ();   /* will be quited by callback function
			of archive (see fileload.c) */
   }

   if (file_exists (temp_file))
      return g_strdup (temp_file);
   else
      return NULL;
}


gboolean
image_info_rename_image (ImageInfo *info, const gchar *filename)
{
   struct stat st;
   gboolean success;

   g_return_val_if_fail (info, FALSE);
   g_return_val_if_fail (info->filename, FALSE);
   g_return_val_if_fail (!info->archive, FALSE);
   g_return_val_if_fail (filename, FALSE);

   success = !rename(info->filename, filename);
   if (success) {
      if (stat (filename, &st)) return FALSE;
      memcpy (&info->st, &st, sizeof (struct stat));
      g_free (info->filename);
      info->filename = (gchar *) g_strdup (filename);
   }

   return success;
}
