/*
 * GImageView
 * Copyright (C) 2001 Takuro Ashie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __IMAGE_VIEW_H__
#define __IMAGE_VIEW_H__

#include "gimageview.h"

#include "fileload.h"
#include "gimv_image.h"
#include "image_info.h"

#define IMAGEVIEW(obj)         GTK_CHECK_CAST (obj, imageview_get_type (), ImageView)
#define IMAGEVIEW_CLASS(klass) GTK_CHECK_CLASS_CAST (klass, imageview_get_type, ImageViewClass)
#define IS_IMAGEVIEW(obj)      GTK_CHECK_TYPE (obj, imageview_get_type ())

#define IMG_MIN_SCALE 5
#define IMG_MAX_SCALE 100000
#define IMG_SCALE_STEP 5

typedef enum
{
   ZOOM_100,  /* Real Size */
   ZOOM_IN,   /* Increase Image Size 5% */
   ZOOM_OUT,  /* Decrease Image Size 5% */
   ZOOM_FIT,  /* Fit to Widnow */
   ZOOM_10,
   ZOOM_25,
   ZOOM_50,
   ZOOM_75,
   ZOOM_125,
   ZOOM_150,
   ZOOM_175,
   ZOOM_200,
   ZOOM_300
} ImgZoomType;


typedef enum
{
   ROTATE_0,
   ROTATE_90,
   ROTATE_180,
   ROTATE_270
} ImageAngle;


typedef struct ImageViewClass_Tag ImageViewClass;


typedef gint (*ImageViewButtonCbFunc) (ImageView      *iv,
				       GdkEventButton *event,
				       gpointer        data);

typedef GList *(*ImageViewNextFn)     (ImageView *iv,
				       gpointer   list_owner,
				       GList     *list,
				       gpointer   data);
typedef GList *(*ImageViewPrevFn)     (ImageView *iv,
				       gpointer   list_owner,
				       GList     *list,
				       gpointer   data);
typedef GList *(*ImageViewNthFn)      (ImageView *iv,
				       gpointer   list_owner,
				       GList     *list,
				       guint      nth,
				       gpointer   data);
typedef void (*ImageViewRemoveListFn) (ImageView *iv,
				       gpointer list_owner,
				       gpointer data);

struct ImageView_Tag
{
   GtkVBox          parent;

   /* widget */
   GtkWidget       *draw_area;
   GtkWidget       *hscrollbar;
   GtkWidget       *vscrollbar;
   GtkWidget       *nav_button;
   GtkWidget       *imageview_popup;
   GtkWidget       *zoom_menu;
   GtkWidget       *rotate_menu;
   GdkCursor       *cursor;
   GtkAdjustment   *hadj, *vadj;

   /* information about image */
   ImageInfo       *info;

   /* image */
   GimvImage       *image;
   GdkPixmap       *gdk_pixmap;
   GdkBitmap       *mask;

   GList           *image_list;      /* shuld be accessed by "image list functions" */
   GList           *image_current;   /* same with above */
   gpointer         list_owner;

   GdkGC           *bg_gc;
   GdkColor        *bg_color;

   /* imageview status */
   gboolean         show_scrollbar;

   /* image status */
   gfloat           x_scale;
   gfloat           y_scale;
   gint             width;
   gint             height;
   gint             rotate;
   gboolean         fit_to_frame_when_change_file;
   gboolean         fit_to_frame;
   gboolean         keep_aspect;
   gboolean         buffer;

   /* information about image dragging */
   guint            button;
   gboolean         pressed;
   gboolean         dragging;
   gint             drag_startx;
   gint             drag_starty;
   gint             x_pos;
   gint             y_pos;
   gint             x_pos_drag_start;
   gint             y_pos_drag_start;

   /* functions for parent widget */
   ImageViewButtonCbFunc button_pressed_cb_func;
   gpointer              button_pressed_cb_data;
   ImageViewButtonCbFunc button_released_cb_func;
   gpointer              button_released_cb_data;

   /* same with released except checking whether dragging or not*/
   ImageViewButtonCbFunc button_clicked_cb_func;
   gpointer              button_clicked_cb_data;

   /* image list functions */
   ImageViewNextFn next_fn;
   gpointer        next_fn_data;
   ImageViewPrevFn prev_fn;
   gpointer        prev_fn_data;
   ImageViewNthFn  nth_fn;
   gpointer        nth_fn_data;

   ImageViewRemoveListFn remove_list_fn;
   gpointer              remove_list_fn_data;
};


struct ImageViewClass_Tag
{
   GtkVBoxClass parent_class;

   void (*set_list)      (ImageView *iv);
   void (*unset_list)    (ImageView *iv);
   void (*rendered)      (ImageView *iv);
   void (*toggle_aspect) (ImageView *iv, gboolean keep_aspect);
   void (*toggle_buffer) (ImageView *iv, gboolean buffered);
};


extern GList *ImageViewList;


guint      imageview_get_type             ();
GtkWidget *imageview_new                  (const gchar *filename);
GtkWidget *imageview_create_zoom_menu     (GtkWidget *window,
					   ImageView *iv);
GtkWidget *imageview_create_rotate_menu   (GtkWidget *window,
					   ImageView *iv);
GtkWidget *imageview_create_popup_menu    (GtkWidget *window,
					   ImageView *iv);

void       imageview_free                 (ImageView   *iv);
void       imageview_free_image_buf       (ImageView   *iv);
void       imageview_load_image_buf       (ImageView   *iv);

void       imageview_change_image_file    (ImageView   *iv,
					   const gchar *newfile);
void       imageview_zoom                 (ImageView   *iv,
					   ImgZoomType  zoom,
					   gint         x_scale,
					   gint         y_scale);
void       imageview_draw_image           (ImageView   *iv);
void       imageview_show_image           (ImageView   *iv,
					   ImageAngle   angle);
void       imageview_rotate_ccw           (ImageView   *iv);
void       imageview_rotate_cw            (ImageView   *iv);
void       imageview_moveto               (ImageView    *iv,
					   gint          x,
					   gint          y);

void       imageview_set_list             (ImageView       *iv,
					   GList           *list,
					   GList           *current,
					   gpointer         list_owner,
					   ImageViewNextFn  next_fn,
					   gpointer         next_fn_data,
					   ImageViewPrevFn  prev_fn,
					   gpointer         prev_fn_data,
					   ImageViewNthFn   nth_fn,
					   gpointer         nth_fn_data,
					   ImageViewRemoveListFn remove_list_fn,
					   gpointer              remove_list_fn_data);
void       imageview_remove_list          (ImageView      *iv,
					   gpointer        list_owner);
void       imageview_popup_menu           (ImageView      *iv,
					   GdkEventButton *event);
void       imageview_next                 (ImageView      *iv);
void       imageview_prev                 (ImageView      *iv);
void       imageview_nth                  (ImageView      *iv,
					   guint           nth);
void       imageview_set_bg_color         (ImageView      *iv,
					   GdkColor       *color);
void       imageview_show_scrollbar       (ImageView      *iv);
void       imageview_hide_scrollbar       (ImageView      *iv);
void       imageview_get_image_frame_size (ImageView      *iv,
					   gint           *width,
					   gint           *height);
void       imageview_open_navwin          (ImageView      *iv,
					   gfloat          x_root,
					   gfloat          y_root);


#endif /* __IMAGE_VIEW_H__ */
