/*
 * GImageView
 * Copyright (C) 2001 Takuro Ashie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>

#include "gimageview.h"

#include "gtkutils.h"
#include "gtk2-compat.h"
#include "menu.h"
#include "prefs.h"
#include "prefs_win.h"
#include "thumbnail_view.h"
#include "thumbnail_window.h"
#include "thumbview_list.h"
#include "thumbview_detail.h"


typedef struct PrefsWin_Tag {
   /* thumbnail window */
   GtkWidget *thumbwin_width_spin;
   GtkWidget *thumbwin_height_spin;
   GtkWidget *thumbwin_layout;
   GtkWidget *thumbwin_dirview_toggle;
   GtkWidget *thumbwin_preview_toggle;
   GtkWidget *thumbwin_preview_tab_toggle;
   GtkWidget *thumbwin_menubar_toggle;
   GtkWidget *thumbwin_toolbar_toggle;
   GtkWidget *thumbwin_statusbar_toggle;
   GtkWidget *thumbwin_tab_toggle;
   GtkWidget *thumbwin_tab_pos;

   /* thumbnail view */
   GtkWidget *thumbview_display_mode;
   GtkWidget *thumbview_thumb_size;

   /* list view */
   GtkWidget *listview_possible;
   GtkWidget *listview_displayed;
   GtkWidget *listview_add;
   GtkWidget *listview_del;
   GtkWidget *listview_up;
   GtkWidget *listview_down;
   gint       listview_possible_selected;
   gint       listview_displayed_selected;

   /* detail view */
   GtkWidget *detail_possible;
   GtkWidget *detail_displayed;
   GtkWidget *detail_add;
   GtkWidget *detail_del;
   GtkWidget *detail_up;
   GtkWidget *detail_down;
   gint       detail_possible_selected;
   gint       detail_displayed_selected;

   /* Directory View */
   GtkWidget *dirview_show_current_dir;
   GtkWidget *dirview_show_parent_dir;

   /* Preview */
   GtkWidget *preview_scale_spin;
   GtkWidget *preview_scrollbar_toggle;
} PrefsWin;


static const gchar *layout_items[] = {
   N_("Layout0"),
   N_("Layout1"),
   N_("Layout2"),
   N_("Layout3"),
   N_("Layout4"),
   N_("Custom"),
   NULL
};


static const gchar *toolbar_style_items[] = {
   N_("Icon"),
   N_("Text"),
   N_("Both"),
   NULL
};


static const gchar *tab_pos_items[] = {
   N_("Left"),
   N_("Right"),
   N_("Top"),
   N_("Bottom"),
   NULL
};


static const gchar *ctree_line_style_items[] = {
   N_("None"),
   N_("Solid"),
   N_("Dotted"),
   N_("Tabbed"),
   NULL
};


static const gchar *ctree_expander_style_items[] = {
   N_("None"),
   N_("Square"),
   N_("Triangle"),
   N_("Circular"),
   NULL
};


static const gchar *thumbview_mouse_items[] = {
   N_("None"),
   N_("Popup menu"),
   N_("Open (Auto select window)"),
   N_("Open (Auto & Force)"),
   N_("Open in preview"),
   N_("Open in preview (Force)"),
   N_("Open in new window"),
   N_("Open in shared window"),
   N_("Open in shared window (Force)"),
   NULL
};


static const gchar *preview_mouse_items[] = {
   N_("None"),
   N_("Next image"),
   N_("Previous image"),
   N_("Open in shared window"),
   N_("Open in new window"),
   N_("Popup menu"),
   N_("Zoom in"),
   N_("Zoom out"),
   N_("Fit image size to frame"),
   N_("Rotate CCW"),
   N_("Rotate CW"),
   N_("Open navigate window"),
   NULL
};


static PrefsWin  prefs_win;
extern Config   *config_changed;
extern Config   *config_prechanged;


static void
set_sensitive_thumbwin_state ()
{
   GtkWidget *widgets[] = {
      prefs_win.thumbwin_width_spin,
      prefs_win.thumbwin_height_spin,
      prefs_win.thumbwin_layout,
      prefs_win.thumbwin_dirview_toggle,
      prefs_win.thumbwin_preview_toggle,
      prefs_win.thumbwin_preview_tab_toggle,
      prefs_win.thumbwin_menubar_toggle,
      prefs_win.thumbwin_toolbar_toggle,
      prefs_win.thumbwin_statusbar_toggle,
      prefs_win.thumbwin_tab_toggle,
      prefs_win.thumbwin_tab_pos,
      prefs_win.thumbview_display_mode,
      prefs_win.thumbview_thumb_size,
      prefs_win.preview_scrollbar_toggle,
   };
   gint i, num = sizeof (widgets) / sizeof (widgets[0]);

   for (i = 0; i < num; i++)
      gtk_widget_set_sensitive (widgets[i],
				!config_changed->thumbwin_save_win_state);
}


static void
cb_rotate_min_colnum_value (GtkWidget *widget, GtkWidget *spinner)
{
   gint colmin;

   colmin = GTK_ADJUSTMENT(widget)->value;

   if (colmin > config_changed->thumbtable_colnum_max) {
      config_changed->thumbtable_colnum_min = config_changed->thumbtable_colnum_max; 
      gtk_spin_button_set_value (GTK_SPIN_BUTTON(spinner),
				 (gfloat) config_changed->thumbtable_colnum_min);
   } else {
      config_changed->thumbtable_colnum_min = colmin; 
   }
}


static void
cb_rotate_max_colnum_value (GtkWidget *widget, GtkWidget *spinner)
{
   gint colmax;

   colmax = GTK_ADJUSTMENT(widget)->value;

   if (colmax < config_changed->thumbtable_colnum_min) {
      config_changed->thumbtable_colnum_max = config_changed->thumbtable_colnum_min; 
      gtk_spin_button_set_value (GTK_SPIN_BUTTON(spinner),
				 (gfloat) config_changed->thumbtable_colnum_max);
   } else {
      config_changed->thumbtable_colnum_max = colmax; 
   }
}


static void
cb_save_thumbwin_state (GtkWidget *toggle)
{
   config_changed->thumbwin_save_win_state =
      gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(toggle));

   set_sensitive_thumbwin_state ();
}


static void
cb_default_disp_mode (GtkWidget *widget)
{
   gchar *tmpstr;
   gint disp_mode;

   disp_mode = GPOINTER_TO_INT (gtk_object_get_data (GTK_OBJECT (widget), "num"));
   tmpstr = thumbview_num_to_label (disp_mode);
   if (tmpstr)
      config_changed->thumbwin_disp_mode = tmpstr;
   else
      config_changed->thumbwin_disp_mode = DEFAULT_DISP_MODE;
}


static void
detail_column_add (gint idx)
{
   gint row, row_possible;
   gchar *text;

   text = _(detailview_get_title (idx));
   if (!text) return;

   row_possible = gtk_clist_find_row_from_data (GTK_CLIST (prefs_win.detail_possible),
						GINT_TO_POINTER (idx));
   if (row_possible < 0) return;

   row = gtk_clist_append (GTK_CLIST (prefs_win.detail_displayed), &text);
   gtk_clist_set_row_data (GTK_CLIST (prefs_win.detail_displayed),
			   row, GINT_TO_POINTER (idx));
   gtk_clist_remove (GTK_CLIST (prefs_win.detail_possible),
		     row_possible);

}


static void
detail_column_del (gint idx)
{
   gint row, row_possible;
   gchar *text;
   gpointer rowdata;

   text = _(detailview_get_title (idx));
   if (!text) return;

   row_possible = gtk_clist_find_row_from_data (GTK_CLIST (prefs_win.detail_displayed),
						GINT_TO_POINTER (idx));
   if (row_possible < 0) return;

   /* remove item from right side */
   gtk_clist_freeze (GTK_CLIST (prefs_win.detail_possible));

   row = gtk_clist_append (GTK_CLIST (prefs_win.detail_possible), &text);
   gtk_clist_set_row_data (GTK_CLIST (prefs_win.detail_possible),
			   row, GINT_TO_POINTER (idx));
   while (row > 0) {
      gint idx_prev;

      rowdata = gtk_clist_get_row_data (GTK_CLIST (prefs_win.detail_possible),
					row - 1);
      if (!rowdata) break;

      idx_prev = GPOINTER_TO_INT (rowdata);

      if (idx < idx_prev) {
	 gtk_clist_swap_rows (GTK_CLIST (prefs_win.detail_possible),
			      row, row - 1);
	 row--;
      } else {
	 break;
      }
   }

   gtk_clist_thaw (GTK_CLIST (prefs_win.detail_possible));

   /* append item to left side */
   gtk_clist_remove (GTK_CLIST (prefs_win.detail_displayed),
		     row_possible);

}


static void
detail_set_data (gchar *string)
{
   gchar **titles;
   gint i, idx;

   if (!string) return;

   titles = g_strsplit (string, ",", -1);

   for (i = 0; titles[i]; i++) {
      idx = detailview_get_title_idx (titles[i]);
      if (idx > 0)
	 detail_column_add (idx);
   }

   g_strfreev (titles);
}


static void
detail_update_config ()
{
   gint rownum = GTK_CLIST (prefs_win.detail_displayed)->rows;
   gint i, idx;
   gpointer rowdata;
   gchar *string = NULL, *tmpstr;

   for (i = 0; i < rownum; i++) {
      rowdata = gtk_clist_get_row_data (GTK_CLIST (prefs_win.detail_displayed), i);
      if (!rowdata) return;

      idx = GPOINTER_TO_INT (rowdata);

      if (string) {
	 tmpstr = string;
	 string = g_strconcat (tmpstr, ",", NULL);
	 g_free (tmpstr);
      } else {
	 string = g_strdup ("");
      }
      tmpstr = string;
      string = g_strconcat (tmpstr, detailview_get_title (idx), NULL);
      g_free (tmpstr);
   }

   if (config_changed->detailview_columns
       != config_prechanged->detailview_columns)
   {
      g_free (config_changed->detailview_columns);
   }
   config_changed->detailview_columns = string;
}


static void
set_sensitive_detail_view ()
{
   GtkWidget *clist1 = prefs_win.detail_possible;
   GtkWidget *clist2 = prefs_win.detail_displayed;
   gint selected1 = prefs_win.detail_possible_selected;
   gint selected2 = prefs_win.detail_displayed_selected;
   gint rownum1, rownum2;

   g_return_if_fail (clist1 && clist2);

   rownum1 = GTK_CLIST (clist1)->rows;
   rownum2 = GTK_CLIST (clist2)->rows;

   /* add button */
   if (selected1 < 0 || selected1 >= rownum1) {
      gtk_widget_set_sensitive (prefs_win.detail_add, FALSE);
   } else {
      gtk_widget_set_sensitive (prefs_win.detail_add, TRUE);
   }

   /* delete button */
   if (selected2 < 0 || selected2 >= rownum2) {
      gtk_widget_set_sensitive (prefs_win.detail_del, FALSE);
   } else {
      gtk_widget_set_sensitive (prefs_win.detail_del, TRUE);
   }

   /* up button */
   if (selected2 > 0 && selected2 < rownum2) {
      gtk_widget_set_sensitive (prefs_win.detail_up, TRUE);
   } else {
      gtk_widget_set_sensitive (prefs_win.detail_up, FALSE);
   }

   /* down button */
   if (selected2 >= 0 && selected2 < rownum2 - 1) {
      gtk_widget_set_sensitive (prefs_win.detail_down, TRUE);
   } else {
      gtk_widget_set_sensitive (prefs_win.detail_down, FALSE);
   }
}


static void
cb_detail_clist_select_row (GtkCList *clist, gint row, gint col,
			    GdkEventButton *event, gpointer data)
{
   gint *selected = data;

   g_return_if_fail (clist);
   g_return_if_fail (data);

   *selected = row;

   set_sensitive_detail_view ();
}


static void
cb_detail_clist_unselect_row (GtkCList *clist, gint row, gint col,
			      GdkEventButton *event, gpointer data)
{
   gint *selected = data;

   g_return_if_fail (clist);
   g_return_if_fail (data);

   *selected = -1;

   set_sensitive_detail_view ();
}


static void
cb_detail_add_button_pressed (GtkButton *button, gpointer data)
{
   gint idx;
   gpointer rowdata;

   if (prefs_win.detail_possible_selected < 0) return;

   rowdata = gtk_clist_get_row_data (GTK_CLIST (prefs_win.detail_possible),
				     prefs_win.detail_possible_selected);
   if (!rowdata) return;

   idx = GPOINTER_TO_INT (rowdata);
   detail_column_add (idx);

   detail_update_config ();

   set_sensitive_detail_view ();
}


static void
cb_detail_del_button_pressed (GtkButton *button, gpointer data)
{
   gint idx;
   gpointer rowdata;

   if (prefs_win.detail_displayed_selected < 0) return;

   rowdata = gtk_clist_get_row_data (GTK_CLIST (prefs_win.detail_displayed),
				     prefs_win.detail_displayed_selected);
   if (!rowdata) return;

   idx = GPOINTER_TO_INT (rowdata);
   detail_column_del (idx);

   detail_update_config ();

   set_sensitive_detail_view ();
}


static void
cb_detail_up_button_pressed (GtkButton *button, gpointer data)
{
   GtkCList *clist = data;
   gint selected = prefs_win.detail_displayed_selected;
   gint num = detailview_get_titles_num () - 1;
   gint rownum = GTK_CLIST (prefs_win.detail_displayed)->rows;

   g_return_if_fail (button && clist);

   if (selected < 1 || selected > rownum - 1
       || selected > num - 1)
   {
      return;
   }
   gtk_clist_swap_rows (clist, selected, selected - 1);
   prefs_win.detail_displayed_selected--;
   gtk_clist_moveto (clist, selected - 1, 0, 0, 0);

   detail_update_config ();

   set_sensitive_detail_view ();
}


static void
cb_detail_down_button_pressed (GtkButton *button, gpointer data)
{
   GtkCList *clist = data;
   gint selected = prefs_win.detail_displayed_selected;
   gint num = detailview_get_titles_num () - 1;
   gint rownum = GTK_CLIST (prefs_win.detail_displayed)->rows;

   g_return_if_fail (button && clist);

   if (selected < 0 || selected > rownum - 2
       || selected > num - 2)
   {
      return;
   }

   gtk_clist_swap_rows (clist, selected, selected + 1);
   prefs_win.detail_displayed_selected++;
   gtk_clist_moveto (clist, selected + 1, 0, 0, 0);

   detail_update_config ();

   set_sensitive_detail_view ();
}


static void
listview_column_add (gchar *title)
{
   GtkCList *clist_possible = GTK_CLIST (prefs_win.listview_possible);
   gint i, idx, row, row_possible = -1;
   gchar *text = NULL, *coltext;

   g_return_if_fail (title && *title);

   idx = listview_get_title_idx (title);
   if (idx < 0) return;

   for (i = 0; i < clist_possible->rows; i++) {
      text = gtk_clist_get_row_data (clist_possible, i);
      if (text && *text && !strcmp (title, text)) {
	 coltext = _(text);
	 row_possible = i;
	 break;
      }
   }
   if (row_possible < 0) return;

   row = gtk_clist_append (GTK_CLIST (prefs_win.listview_displayed), &coltext);
   gtk_clist_set_row_data (GTK_CLIST (prefs_win.listview_displayed),
			   row, text);
   gtk_clist_remove (clist_possible, row_possible);
}


static void
listview_column_del (gchar *title)
{
   GtkCList *clist_displayed = GTK_CLIST (prefs_win.listview_displayed);
   gint i, row, row_possible = -1, idx;
   gchar *text = NULL, *coltext;

   g_return_if_fail (title && *title);

   idx = listview_get_title_idx (title);
   if (idx < 0) return;

   for (i = 0; i < clist_displayed->rows; i++) {
      text = gtk_clist_get_row_data (clist_displayed, i);
      if (text && *text && !strcmp (title, text)) {
	 coltext = _(text);
	 row_possible = i;
	 break;
      }
   }

   if (row_possible < 0) return;

   /* remove item from right side */
   gtk_clist_freeze (GTK_CLIST (prefs_win.listview_possible));

   row = gtk_clist_append (GTK_CLIST (prefs_win.listview_possible), &coltext);
   gtk_clist_set_row_data (GTK_CLIST (prefs_win.listview_possible),
			   row, text);
   while (row > 0) {
      gint idx_prev;

      text = gtk_clist_get_row_data (GTK_CLIST (prefs_win.listview_possible),
				     row - 1);
      if (!text) break;

      idx_prev = listview_get_title_idx (text);

      if (idx < idx_prev) {
	 gtk_clist_swap_rows (GTK_CLIST (prefs_win.listview_possible),
			      row, row - 1);
	 row--;
      } else {
	 break;
      }
   }

   gtk_clist_thaw (GTK_CLIST (prefs_win.listview_possible));

   /* append item to left side */
   gtk_clist_remove (clist_displayed,
		     row_possible);

}


static void
listview_data_order_set_data (gchar *string)
{
   gchar **titles;
   gint i;

   if (!string) return;

   titles = g_strsplit (string, ",", -1);

   for (i = 0; titles[i]; i++)
      listview_column_add (titles[i]);

   g_strfreev (titles);
}


static void
listview_update_config ()
{
   gint rownum = GTK_CLIST (prefs_win.listview_displayed)->rows;
   gint i;
   gchar *rowdata;
   gchar *string = NULL, *tmpstr;

   for (i = 0; i < rownum; i++) {
      rowdata = gtk_clist_get_row_data (GTK_CLIST (prefs_win.listview_displayed), i);
      if (!rowdata) return;

      if (string) {
	 tmpstr = string;
	 string = g_strconcat (tmpstr, ",", NULL);
	 g_free (tmpstr);
      } else {
	 string = g_strdup ("");
      }
      tmpstr = string;
      string = g_strconcat (tmpstr, rowdata, NULL);
      g_free (tmpstr);
   }

   if (config_changed->listview_data_order
       != config_prechanged->listview_data_order)
   {
      g_free (config_changed->listview_data_order);
   }
   config_changed->listview_data_order = string;
}


static void
set_sensitive_list_view ()
{
   GtkWidget *clist1 = prefs_win.listview_possible;
   GtkWidget *clist2 = prefs_win.listview_displayed;
   gint selected1 = prefs_win.listview_possible_selected;
   gint selected2 = prefs_win.listview_displayed_selected;
   gint rownum1, rownum2;

   g_return_if_fail (clist1 && clist2);

   rownum1 = GTK_CLIST (clist1)->rows;
   rownum2 = GTK_CLIST (clist2)->rows;

   /* add button */
   if (selected1 < 0 || selected1 >= rownum1) {
      gtk_widget_set_sensitive (prefs_win.listview_add, FALSE);
   } else {
      gtk_widget_set_sensitive (prefs_win.listview_add, TRUE);
   }

   /* delete button */
   if (selected2 < 0 || selected2 >= rownum2) {
      gtk_widget_set_sensitive (prefs_win.listview_del, FALSE);
   } else {
      gtk_widget_set_sensitive (prefs_win.listview_del, TRUE);
   }

   /* up button */
   if (selected2 > 0 && selected2 < rownum2) {
      gtk_widget_set_sensitive (prefs_win.listview_up, TRUE);
   } else {
      gtk_widget_set_sensitive (prefs_win.listview_up, FALSE);
   }

   /* down button */
   if (selected2 >= 0 && selected2 < rownum2 - 1) {
      gtk_widget_set_sensitive (prefs_win.listview_down, TRUE);
   } else {
      gtk_widget_set_sensitive (prefs_win.listview_down, FALSE);
   }
}


static void
cb_listview_clist_select_row (GtkCList *clist, gint row, gint col,
			      GdkEventButton *event, gpointer data)
{
   gint *selected = data;

   g_return_if_fail (clist);
   g_return_if_fail (data);

   *selected = row;

   set_sensitive_list_view ();
}


static void
cb_listview_clist_unselect_row (GtkCList *clist, gint row, gint col,
				GdkEventButton *event, gpointer data)
{
   gint *selected = data;

   g_return_if_fail (clist);
   g_return_if_fail (data);

   *selected = -1;

   set_sensitive_list_view ();
}


static void
cb_listview_add_button_pressed (GtkButton *button, gpointer data)
{
   gchar *rowdata;

   if (prefs_win.listview_possible_selected < 0) return;

   rowdata = gtk_clist_get_row_data (GTK_CLIST (prefs_win.listview_possible),
				     prefs_win.listview_possible_selected);
   if (!rowdata) return;

   listview_column_add (rowdata);

   listview_update_config ();

   set_sensitive_list_view ();
}


static void
cb_listview_del_button_pressed (GtkButton *button, gpointer data)
{
   gchar *rowdata;

   if (prefs_win.listview_displayed_selected < 0) return;

   rowdata = gtk_clist_get_row_data (GTK_CLIST (prefs_win.listview_displayed),
				     prefs_win.listview_displayed_selected);
   if (!rowdata) return;

   listview_column_del (rowdata);

   listview_update_config ();

   set_sensitive_list_view ();
}


static void
cb_listview_up_button_pressed (GtkButton *button, gpointer data)
{
   GtkCList *clist = data;
   gint selected = prefs_win.listview_displayed_selected;
   gint num = listview_get_titles_num () - 1;
   gint rownum = GTK_CLIST (prefs_win.listview_displayed)->rows;

   g_return_if_fail (button && clist);

   if (selected < 1 || selected > rownum - 1
       || selected > num - 1)
   {
      return;
   }
   gtk_clist_swap_rows (clist, selected, selected - 1);
   prefs_win.listview_displayed_selected--;
   gtk_clist_moveto (clist, selected - 1, 0, 0, 0);

   listview_update_config ();

   set_sensitive_list_view ();
}


static void
cb_listview_down_button_pressed (GtkButton *button, gpointer data)
{
   GtkCList *clist = data;
   gint selected = prefs_win.listview_displayed_selected;
   gint num = listview_get_titles_num () - 1;
   gint rownum = GTK_CLIST (prefs_win.listview_displayed)->rows;

   g_return_if_fail (button && clist);

   if (selected < 0 || selected > rownum - 2
       || selected > num - 2)
   {
      return;
   }

   gtk_clist_swap_rows (clist, selected, selected + 1);
   prefs_win.listview_displayed_selected++;
   gtk_clist_moveto (clist, selected + 1, 0, 0, 0);

   listview_update_config ();

   set_sensitive_list_view ();
}


static void
set_sensitive_preview ()
{
   gtk_widget_set_sensitive (prefs_win.preview_scale_spin,
			     !config_changed->preview_fit_image_to_frame);
}


static void
cb_fit_image_to_frame (GtkWidget *toggle)
{
   config_changed->preview_fit_image_to_frame =
      gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(toggle));
   set_sensitive_preview ();
}



/*******************************************************************************
 *  prefs_thumbwin_page:
 *     @ Create Thumbnail View preference page.
 *
 *  Return : Packed widget (GtkVbox)
 *******************************************************************************/
GtkWidget *
prefs_thumbwin_page ()
{
   GtkWidget *main_vbox;
   GtkWidget *frame, *vbox, *vbox2, *hbox, *hbox2;
   GtkWidget *label;
   GtkAdjustment *adj;
   GtkWidget *spinner;
   GtkWidget *option_menu;
   GtkWidget *toggle;

   main_vbox = gtk_vbox_new (FALSE, 0);
   gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 5);

   /**********************************************
    * Window Frame
    **********************************************/
   prefs_create_frame (_("Window"), frame, vbox, main_vbox);

   /* Save window state on exit */
   toggle = create_check_button (_("Save window state on exit"),
				 conf.thumbwin_save_win_state,
                                 cb_save_thumbwin_state,
                                 &config_changed->thumbwin_save_win_state);
   gtk_box_pack_start (GTK_BOX (vbox), toggle, FALSE, FALSE, 0);

   /* Window width and height spinner */
   hbox = gtk_hbox_new (FALSE, 10);
   gtk_container_set_border_width (GTK_CONTAINER(hbox), 5);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
   label = gtk_label_new (_("width"));
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   adj = (GtkAdjustment *) gtk_adjustment_new (conf.thumbwin_width,
					       1.0, 10000.0, 1.0, 5.0, 0.0);
   spinner = prefs_win.thumbwin_width_spin = gtk_spin_button_new (adj, 0, 0);
   gtk_widget_set_usize(spinner, 50, -1);
   gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (spinner), TRUE);
   gtk_spin_button_set_shadow_type (GTK_SPIN_BUTTON (spinner), GTK_SHADOW_OUT);
   gtk_signal_connect (GTK_OBJECT (adj), "value_changed",
		       GTK_SIGNAL_FUNC (cb_get_data_from_adjustment_by_int),
		       &config_changed->thumbwin_width);
   gtk_box_pack_start (GTK_BOX (hbox), spinner, FALSE, FALSE, 0);

   label = gtk_label_new (_("height"));
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   adj = (GtkAdjustment *) gtk_adjustment_new (conf.thumbwin_height,
					       1.0, 10000.0, 1.0, 5.0, 0.0);
   spinner = prefs_win.thumbwin_height_spin = gtk_spin_button_new (adj, 0, 0);
   gtk_widget_set_usize(spinner, 50, -1);
   gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (spinner), TRUE);
   gtk_spin_button_set_shadow_type (GTK_SPIN_BUTTON (spinner), GTK_SHADOW_OUT);
   gtk_signal_connect (GTK_OBJECT (adj), "value_changed",
		       GTK_SIGNAL_FUNC (cb_get_data_from_adjustment_by_int),
		       &config_changed->thumbwin_height);
   gtk_box_pack_start (GTK_BOX (hbox), spinner, FALSE, FALSE, 0);

   /* Toolbar Style */
   label = gtk_label_new (_("Default Layout"));
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 5);

   option_menu = create_option_menu_simple (layout_items,
					    conf.thumbwin_layout_type,
					    (gint *) &config_changed->thumbwin_layout_type);
   prefs_win.thumbwin_layout = option_menu;
   gtk_box_pack_start (GTK_BOX (hbox), option_menu, FALSE, FALSE, 0);

   /* Show/Hide frame */
   prefs_create_frame (_("Show/Hide"), frame, vbox2, vbox);

   hbox = gtk_hbox_new (FALSE, 10);
   gtk_container_set_border_width (GTK_CONTAINER(hbox), 5);
   gtk_box_pack_start (GTK_BOX (vbox2), hbox, FALSE, FALSE, 0);

   hbox2 = gtk_hbox_new (FALSE, 10);
   gtk_container_set_border_width (GTK_CONTAINER(hbox), 5);
   gtk_box_pack_start (GTK_BOX (vbox2), hbox2, FALSE, FALSE, 0);

   /* Show Directory or not */
   toggle = create_check_button (_("Directory view"), conf.thumbwin_show_dir_view,
                                 cb_get_data_from_toggle,
                                 &config_changed->thumbwin_show_dir_view);
   prefs_win.thumbwin_dirview_toggle = toggle;
   gtk_box_pack_start (GTK_BOX (hbox), toggle, FALSE, FALSE, 0);

   /* Show Preview or not */
   toggle = create_check_button (_("Preview"), conf.thumbwin_show_preview,
                                 cb_get_data_from_toggle,
                                 &config_changed->thumbwin_show_preview);
   prefs_win.thumbwin_preview_toggle = toggle;
   gtk_box_pack_start (GTK_BOX (hbox), toggle, FALSE, FALSE, 0);

   /* Show Preview Tab or not */
   toggle = create_check_button (_("Preview Tab"), conf.thumbwin_show_preview_tab,
                                 cb_get_data_from_toggle,
                                 &config_changed->thumbwin_show_preview_tab);
   prefs_win.thumbwin_preview_tab_toggle = toggle;
   gtk_box_pack_start (GTK_BOX (hbox), toggle, FALSE, FALSE, 0);

   /* Show Menubar or not */
   toggle = create_check_button (_("Menubar"), conf.thumbwin_show_menubar,
                                 cb_get_data_from_toggle,
                                 &config_changed->thumbwin_show_menubar);
   prefs_win.thumbwin_menubar_toggle = toggle;
   gtk_box_pack_start (GTK_BOX (hbox2), toggle, FALSE, FALSE, 0);

   /* Show Toolbar or not */
   toggle = create_check_button (_("Toolbar"), conf.thumbwin_show_toolbar,
				 cb_get_data_from_toggle,
				 &config_changed->thumbwin_show_toolbar);
   prefs_win.thumbwin_toolbar_toggle = toggle;
   gtk_box_pack_start (GTK_BOX (hbox2), toggle, FALSE, FALSE, 0);

   /* Show Statusbar or not */
   toggle = create_check_button (_("Statusbar"), conf.thumbwin_show_statusbar,
				 cb_get_data_from_toggle,
				 &config_changed->thumbwin_show_statusbar);
   prefs_win.thumbwin_statusbar_toggle = toggle;
   gtk_box_pack_start (GTK_BOX (hbox2), toggle, FALSE, FALSE, 0);

   /* Show Tab or not */
   toggle = create_check_button (_("Tab"), conf.thumbwin_show_tab,
				 cb_get_data_from_toggle,
				 &config_changed->thumbwin_show_tab);
   prefs_win.thumbwin_tab_toggle = toggle;
   gtk_box_pack_start (GTK_BOX (hbox2), toggle, FALSE, FALSE, 0);

   /* Raise window or not */
   toggle = create_check_button (_("Raise window when opening thumbnails"),
				 conf.thumbwin_raise_window,
				 cb_get_data_from_toggle,
				 &config_changed->thumbwin_raise_window);
   gtk_box_pack_start (GTK_BOX (vbox), toggle, FALSE, FALSE, 0);

   hbox = gtk_hbox_new (FALSE, 0);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

   /* Toolbar Style */
   hbox = gtk_hbox_new (FALSE, 0);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

   label = gtk_label_new (_("Toolbar Style"));
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 5);

   option_menu = create_option_menu_simple (toolbar_style_items,
					    conf.thumbwin_toolbar_style,
					    (gint *) &config_changed->thumbwin_toolbar_style);
   gtk_box_pack_start (GTK_BOX (hbox), option_menu, FALSE, FALSE, 5);


   /* GUI redraw interval */
   hbox = gtk_hbox_new (FALSE, 10);
   gtk_container_set_border_width (GTK_CONTAINER (hbox), 5);
   gtk_box_pack_start (GTK_BOX (main_vbox), hbox, FALSE, FALSE, 0);

   label = gtk_label_new (_("GUI redraw interval while loading: Every"));
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   adj = (GtkAdjustment *) gtk_adjustment_new (conf.thumbwin_redraw_interval,
					       1.0, 1000.0, 1.0, 5.0, 0.0);
   spinner = gtk_spin_button_new (adj, 0, 0);
   gtk_widget_set_usize(spinner, 50, -1);
   gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (spinner), TRUE);
   gtk_spin_button_set_shadow_type (GTK_SPIN_BUTTON (spinner), GTK_SHADOW_OUT);
   gtk_signal_connect (GTK_OBJECT (adj), "value_changed",
		       GTK_SIGNAL_FUNC (cb_get_data_from_adjustment_by_int),
		       &config_changed->thumbwin_redraw_interval);
   gtk_box_pack_start (GTK_BOX (hbox), spinner, FALSE, FALSE, 0);

   label = gtk_label_new (_("files"));
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);

   /* set sensitive */
   /* set_sensitive_thumbwin_state (); */

   return main_vbox;
}


/*******************************************************************************
 *  prefs_listview_page:
 *     @ Create Detail View preference page.
 *
 *  Return : Packed widget (GtkVbox)
 *******************************************************************************/
GtkWidget *
prefs_thumbview_toolbar_page ()
{
   GtkWidget *main_vbox;
   GtkWidget *label;

   main_vbox = gtk_vbox_new (FALSE, 0);
   gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 5);

   label = gtk_label_new (_("Not implemented yet."));
   gtk_box_pack_start (GTK_BOX (main_vbox), label, FALSE, FALSE, 5);

   return main_vbox;
}


/*******************************************************************************
 *  prefs_thumbwin_tab_page:
 *     @ Create Thumbnail View preference page.
 *
 *  Return : Packed widget (GtkVbox)
 *******************************************************************************/
GtkWidget *
prefs_thumbwin_tab_page ()
{
   GtkWidget *main_vbox;
   GtkWidget *frame, *vbox, *hbox;
   GtkWidget *label;
   GtkWidget *option_menu;
   GtkWidget *toggle;

   main_vbox = gtk_vbox_new (FALSE, 0);
   gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 5);

   /**********************************************
    * Tab Frame
    **********************************************/
   prefs_create_frame (_("Tab"), frame, vbox, main_vbox);

   /* Tab Position Selection */
   hbox = gtk_hbox_new (FALSE, 0);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
   label = gtk_label_new (_("Tab Position"));
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 5);

   option_menu = create_option_menu_simple (tab_pos_items,
					    conf.thumbwin_tab_pos,
					    (gint *) &config_changed->thumbwin_tab_pos);
   gtk_box_pack_start (GTK_BOX (hbox), option_menu, FALSE, FALSE, 5);

   hbox = gtk_hbox_new (FALSE, 0);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
   prefs_win.thumbwin_tab_pos = option_menu;

   /* Move to new tab automatically or not */
   toggle = create_check_button (_("Move to new tab automatically"),
				 conf.thumbwin_move_to_newtab,
                                 cb_get_data_from_toggle,
                                 &config_changed->thumbwin_move_to_newtab);
   gtk_box_pack_start (GTK_BOX (hbox), toggle, FALSE, FALSE, 5);

   /* show tab close button */
   toggle = create_check_button (_("Show close button"),
				 conf.thumbwin_show_tab_close,
                                 cb_get_data_from_toggle,
                                 &config_changed->thumbwin_show_tab_close);
   gtk_box_pack_start (GTK_BOX (hbox), toggle, FALSE, FALSE, 5);

   hbox = gtk_hbox_new (FALSE, 0);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

   /* show full path in tab */
   toggle = create_check_button (_("Show full path"),
				 conf.thumbwin_tab_fullpath,
                                 cb_get_data_from_toggle,
                                 &config_changed->thumbwin_tab_fullpath);
   gtk_box_pack_start (GTK_BOX (hbox), toggle, FALSE, FALSE, 5);

   hbox = gtk_hbox_new (FALSE, 0);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

   /* force open tab */
   toggle = create_check_button (_("Open new tab whether image is exist or not "
				   "in the directory"),
				 conf.thumbwin_force_open_tab,
                                 cb_get_data_from_toggle,
                                 &config_changed->thumbwin_force_open_tab);
   gtk_box_pack_start (GTK_BOX (hbox), toggle, FALSE, FALSE, 5);

   return main_vbox;
}


GtkWidget *
prefs_thumbview_page ()
{
   GtkWidget *main_vbox;
   GtkWidget *frame, *vbox, *hbox;
   GtkWidget *toggle, *spinner, *label, *option_menu;
   GtkAdjustment *adj;
   gint num, disp_mode;
   gchar **disp_mode_labels;

   main_vbox = gtk_vbox_new (FALSE, 0);
   gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 5);

   /**********************************************
    * Show/Hide Frame
    **********************************************/
   prefs_create_frame (_("Show/Hide"), frame, vbox, main_vbox);

   hbox = gtk_hbox_new (FALSE, 0);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

   /* show directory */
   toggle = create_check_button (_("Directory"),
				 conf.thumbview_show_dir,
                                 cb_get_data_from_toggle,
                                 &config_changed->thumbview_show_dir);
   gtk_box_pack_start (GTK_BOX (hbox), toggle, FALSE, FALSE, 5);

   /* show archive */
   toggle = create_check_button (_("Archive"),
				 conf.thumbview_show_archive,
                                 cb_get_data_from_toggle,
                                 &config_changed->thumbview_show_archive);
   gtk_box_pack_start (GTK_BOX (hbox), toggle, FALSE, FALSE, 5);


   /* Default display mode */
   hbox = gtk_hbox_new (FALSE, 0);
   gtk_box_pack_start (GTK_BOX (main_vbox), hbox, FALSE, FALSE, 0);

   disp_mode_labels = thumbview_get_disp_mode_labels (&num);

   label = gtk_label_new (_("Default display mode"));
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 5);

   disp_mode = thumbview_label_to_num (conf.thumbwin_disp_mode);
   if (disp_mode < 0)
      disp_mode = DEFAULT_DISP_MODE_NUM;
   option_menu = create_option_menu ((const gchar **)disp_mode_labels,
				     disp_mode,
				     cb_default_disp_mode,
				     NULL);
   gtk_box_pack_start (GTK_BOX (hbox), option_menu, FALSE, FALSE, 5);
   prefs_win.thumbview_display_mode = option_menu;

   /* Thumbnail size spinner */
   hbox = gtk_hbox_new (FALSE, 10);
   gtk_container_set_border_width (GTK_CONTAINER (hbox), 5);
   gtk_box_pack_start (GTK_BOX (main_vbox), hbox, FALSE, FALSE, 0);

   label = gtk_label_new (_("Thumnail Size"));
   gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, TRUE, 0);

   adj = (GtkAdjustment *) gtk_adjustment_new (conf.thumbwin_thumb_size,
					       MIN_THUMB_SIZE, MAX_THUMB_SIZE,
					       1.0, 5.0, 0.0);
   spinner = gtk_spin_button_new (adj, 0, 0);
   prefs_win.thumbview_thumb_size = spinner;
   gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (spinner), TRUE);
   gtk_spin_button_set_shadow_type (GTK_SPIN_BUTTON (spinner), GTK_SHADOW_OUT);
   gtk_signal_connect (GTK_OBJECT (adj), "value_changed",
		       GTK_SIGNAL_FUNC (cb_get_data_from_adjustment_by_int),
		       &config_changed->thumbwin_thumb_size);
   gtk_box_pack_start(GTK_BOX(hbox), spinner, FALSE, TRUE, 0);

   return main_vbox;
}


/*******************************************************************************
 *  prefs_thumbtable_page:
 *     @ Create Thumbnail Table preference page.
 *
 *  Return : Packed widget (GtkVbox)
 *******************************************************************************/
GtkWidget *
prefs_thumbtable_page ()
{
   GtkWidget *main_vbox;
   GtkWidget *frame, *table, *vbox, *hbox;
   GtkWidget *label;
   GtkAdjustment *adj;
   GtkWidget *spinner;

   main_vbox = gtk_vbox_new (FALSE, 0);
   gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 5);

   /********************************************** 
    * Thumbnail Table Frame
    **********************************************/
   frame = gtk_frame_new (_("Thumbnail Table"));
   gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
   gtk_box_pack_start(GTK_BOX(main_vbox), frame, FALSE, TRUE, 0);

   table = gtk_table_new (2, 4, FALSE);
   gtk_container_set_border_width(GTK_CONTAINER(table), 5);
   gtk_container_add (GTK_CONTAINER (frame), table);

   /* Minimum column number spinner */
   label = gtk_label_new (_("Minimum Columns"));
   gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1,
		     GTK_EXPAND, GTK_FILL, 0, 0);

   adj = (GtkAdjustment *) gtk_adjustment_new (conf.thumbtable_colnum_min,
					       1.0, 256.0, 1.0, 5.0, 0.0);
   spinner = gtk_spin_button_new (adj, 0, 0);
   gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (spinner), TRUE);
   gtk_spin_button_set_shadow_type (GTK_SPIN_BUTTON (spinner), GTK_SHADOW_OUT);
   gtk_signal_connect (GTK_OBJECT (adj), "value_changed",
		       GTK_SIGNAL_FUNC (cb_rotate_min_colnum_value), spinner);
   gtk_table_attach (GTK_TABLE (table), spinner, 1, 2, 0, 1,
		     GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0);

   /* Maximum column number spinner */
   label = gtk_label_new (_("Maximum Columns"));
   gtk_table_attach (GTK_TABLE (table), label, 2, 3, 0, 1,
		     GTK_EXPAND, GTK_FILL, 0, 0);

   adj = (GtkAdjustment *) gtk_adjustment_new (conf.thumbtable_colnum_max,
					       1.0, 256.0, 1.0, 5.0, 0.0);
   spinner = gtk_spin_button_new (adj, 0, 0);
   gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (spinner), TRUE);
   gtk_spin_button_set_shadow_type (GTK_SPIN_BUTTON (spinner), GTK_SHADOW_OUT);
   gtk_signal_connect (GTK_OBJECT (adj), "value_changed",
		       GTK_SIGNAL_FUNC (cb_rotate_max_colnum_value), spinner);
   gtk_table_attach (GTK_TABLE (table), spinner, 3, 4, 0, 1,
		     GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0);

   /* Row Spacing spinner */
   label = gtk_label_new (_("Row Spacing"));
   gtk_table_attach (GTK_TABLE (table), label, 0, 1, 1, 2,
		     GTK_EXPAND, GTK_FILL, 0, 0);

   adj = (GtkAdjustment *) gtk_adjustment_new (conf.thumbtable_row_space,
					       0.0, 255.0, 1.0, 5.0, 0.0);
   spinner = gtk_spin_button_new (adj, 0, 0);
   gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (spinner), TRUE);
   gtk_spin_button_set_shadow_type (GTK_SPIN_BUTTON (spinner), GTK_SHADOW_OUT);
   gtk_signal_connect (GTK_OBJECT (adj), "value_changed",
		       GTK_SIGNAL_FUNC (cb_get_data_from_adjustment_by_int),
		       &config_changed->thumbtable_row_space);
   gtk_table_attach (GTK_TABLE (table), spinner, 1, 2, 1, 2,
		     GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0);

   /* Column Spacing spinner */
   label = gtk_label_new (_("Column Spacing"));
   gtk_table_attach (GTK_TABLE (table), label, 2, 3, 1, 2,
		     GTK_EXPAND, GTK_FILL, 0, 0);

   adj = (GtkAdjustment *) gtk_adjustment_new (conf.thumbtable_col_space,
					       0.0, 255.0, 1.0, 5.0, 0.0);
   spinner = gtk_spin_button_new (adj, 0, 0);
   gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (spinner), TRUE);
   gtk_spin_button_set_shadow_type (GTK_SPIN_BUTTON (spinner), GTK_SHADOW_OUT);
   gtk_signal_connect (GTK_OBJECT (adj), "value_changed",
		       GTK_SIGNAL_FUNC (cb_get_data_from_adjustment_by_int),
		       &config_changed->thumbtable_col_space);
   gtk_table_attach (GTK_TABLE (table), spinner, 3, 4, 1, 2,
		     GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0);

   /**********************************************
    * Thumbnail Button Frame
    **********************************************/
   prefs_create_frame (_("Thumbnail Button"), frame, vbox, main_vbox);

   hbox = gtk_hbox_new (FALSE, 10);
   gtk_container_set_border_width (GTK_CONTAINER(hbox), 5);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

   /* Minimum border width spinner */
   label = gtk_label_new (_("Minimum Border Width"));
   gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, TRUE, 0);

   adj = (GtkAdjustment *) gtk_adjustment_new (conf.thumbbutton_border_width,
					       0.0, 32.0, 1.0, 5.0, 0.0);
   spinner = gtk_spin_button_new (adj, 0, 0);
   gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (spinner), TRUE);
   gtk_spin_button_set_shadow_type (GTK_SPIN_BUTTON (spinner), GTK_SHADOW_OUT);
   gtk_signal_connect (GTK_OBJECT (adj), "value_changed",
		       GTK_SIGNAL_FUNC (cb_get_data_from_adjustment_by_int),
		       &config_changed->thumbbutton_border_width);
   gtk_box_pack_start(GTK_BOX(hbox), spinner, FALSE, TRUE, 0);

   return main_vbox;
}


/*******************************************************************************
 *  prefs_listview_page:
 *     @ Create List View preference page.
 *
 *  Return : Packed widget (GtkVbox)
 *******************************************************************************/
GtkWidget *
prefs_listview_page ()
{
   GtkWidget *main_vbox;
   GtkWidget *frame, *frame_vbox, *vbox, *table;
   GtkWidget *scrollwin, *clist, *button1, *button2;
   GtkWidget *hseparator, *label;
   gint i;

   main_vbox = gtk_vbox_new (FALSE, 0);
   gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 5);

   /**********************************************
    *  Data Order on List(Thumbnail) frame
    **********************************************/
   prefs_create_frame (_("Data Order on List View (Thumbnail)"),
		       frame, frame_vbox, main_vbox);

   /* clist */
   table = gtk_table_new (2, 3, FALSE);
   gtk_container_add (GTK_CONTAINER (frame_vbox), table);

   /* possible columns */
   label = gtk_label_new (_("Possible data"));
   gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1,
		     0, 0, 0, 0);

   scrollwin = gtk_scrolled_window_new (NULL, NULL);
   gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrollwin),
				   GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
   gtk_container_set_border_width(GTK_CONTAINER(scrollwin), 5);
   gtk_table_attach (GTK_TABLE (table), scrollwin, 0, 1, 1, 2,
		     GTK_FILL | GTK_EXPAND, GTK_EXPAND, 0, 0);
   gtk_widget_set_usize (scrollwin, -1, 200);

   clist = prefs_win.listview_possible =  gtk_clist_new (1);
   gtk_clist_set_selection_mode (GTK_CLIST (clist), GTK_SELECTION_SINGLE);
   gtk_clist_set_column_auto_resize (GTK_CLIST (clist), 0, TRUE);
   gtk_container_add (GTK_CONTAINER (scrollwin), clist);

   for (i = 0; i < listview_get_titles_num (); i++) {
      gchar *text = _(listview_get_title (i));
      gtk_clist_append (GTK_CLIST (clist), &text);
      gtk_clist_set_row_data (GTK_CLIST (clist), i, listview_get_title (i));
   }

   gtk_signal_connect (GTK_OBJECT (clist),"select_row",
		       GTK_SIGNAL_FUNC (cb_listview_clist_select_row),
		       &prefs_win.listview_possible_selected);
   gtk_signal_connect (GTK_OBJECT (clist),"unselect_row",
		       GTK_SIGNAL_FUNC (cb_listview_clist_unselect_row),
		       &prefs_win.listview_possible_selected);

   prefs_win.listview_possible_selected = -1;

   /* add/delete buttons */
   vbox = gtk_vbox_new (TRUE, 0);
   gtk_table_attach (GTK_TABLE (table), vbox, 1, 2, 1, 2,
		     GTK_FILL, GTK_FILL, 0, 0);

   button1 = gtk_button_new_with_label (_("Add"));
   prefs_win.listview_add = button1;
   gtk_widget_set_usize (button1, 50, -1);
   gtk_box_pack_start (GTK_BOX (vbox), button1, FALSE, FALSE, 2);

   button2 = gtk_button_new_with_label (_("Delete"));
   prefs_win.listview_del = button2;
   gtk_widget_set_usize (button2, 50, -1);
   gtk_box_pack_start (GTK_BOX (vbox), button2, FALSE, FALSE, 2);

   gtk_signal_connect (GTK_OBJECT (button1),"released",
		       GTK_SIGNAL_FUNC (cb_listview_add_button_pressed), clist);
   gtk_signal_connect (GTK_OBJECT (button2),"released",
		       GTK_SIGNAL_FUNC (cb_listview_del_button_pressed), clist);

   hseparator = gtk_hseparator_new ();
   gtk_box_pack_start (GTK_BOX (vbox), hseparator, FALSE, FALSE, 2);

   /* move buttons */
   button1 = gtk_button_new_with_label (_("Up"));
   prefs_win.listview_up = button1;
   gtk_widget_set_usize (button1, 50, -1);
   gtk_box_pack_start (GTK_BOX (vbox), button1, FALSE, FALSE, 2);

   button2 = gtk_button_new_with_label (_("Down"));
   prefs_win.listview_down = button2;
   gtk_widget_set_usize (button2, 50, -1);
   gtk_box_pack_start (GTK_BOX (vbox), button2, FALSE, FALSE, 2);

   /* displayed columns */
   label = gtk_label_new (_("Displayed data"));
   gtk_table_attach (GTK_TABLE (table), label, 2, 3, 0, 1,
		     0, 0, 0, 0);

   scrollwin = gtk_scrolled_window_new (NULL, NULL);
   gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrollwin),
				   GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
   gtk_container_set_border_width(GTK_CONTAINER(scrollwin), 5);
   gtk_table_attach (GTK_TABLE (table), scrollwin, 2, 3, 1, 2,
		     GTK_FILL | GTK_EXPAND, GTK_EXPAND, 0, 0);
   gtk_widget_set_usize (scrollwin, -1, 200);

   clist = prefs_win.listview_displayed = gtk_clist_new (1);
   gtk_clist_set_selection_mode (GTK_CLIST (clist), GTK_SELECTION_SINGLE);
   gtk_clist_set_column_auto_resize (GTK_CLIST (clist), 0, TRUE);
   gtk_container_add (GTK_CONTAINER (scrollwin), clist);

   gtk_signal_connect (GTK_OBJECT (clist),"select_row",
		       GTK_SIGNAL_FUNC (cb_listview_clist_select_row),
		       &prefs_win.listview_displayed_selected);
   gtk_signal_connect (GTK_OBJECT (clist),"unselect_row",
		       GTK_SIGNAL_FUNC (cb_listview_clist_unselect_row),
		       &prefs_win.listview_displayed_selected);

   prefs_win.listview_displayed_selected = -1;
   listview_data_order_set_data (conf.listview_data_order);

   gtk_signal_connect (GTK_OBJECT (button1),"released",
		       GTK_SIGNAL_FUNC (cb_listview_up_button_pressed), clist);
   gtk_signal_connect (GTK_OBJECT (button2),"released",
		       GTK_SIGNAL_FUNC (cb_listview_down_button_pressed), clist);

   return main_vbox;
}


/*******************************************************************************
 *  prefs_detailview_page:
 *     @ Create Detail View preference page.
 *
 *  Return : Packed widget (GtkVbox)
 *******************************************************************************/
GtkWidget *
prefs_detailview_page ()
{
   GtkWidget *main_vbox;
   GtkWidget *frame, *frame_vbox, *vbox, *table;
   GtkWidget *scrollwin, *clist, *button1, *button2;
   GtkWidget *hseparator, *toggle, *label;
   gint i;

   main_vbox = gtk_vbox_new (FALSE, 0);
   gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 5);

   /**********************************************
    *  Column Order frame
    **********************************************/
   prefs_create_frame (_("Column Order"), frame, frame_vbox, main_vbox);

   /* clist */
   table = gtk_table_new (2, 3, FALSE);
   gtk_container_add (GTK_CONTAINER (frame_vbox), table);

   /* possible columns */
   label = gtk_label_new (_("Possible columns"));
   gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1,
		     0, 0, 0, 0);

   scrollwin = gtk_scrolled_window_new (NULL, NULL);
   gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrollwin),
				   GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
   gtk_container_set_border_width(GTK_CONTAINER(scrollwin), 5);
   gtk_table_attach (GTK_TABLE (table), scrollwin, 0, 1, 1, 2,
		     GTK_FILL | GTK_EXPAND, GTK_EXPAND, 0, 0);
   gtk_widget_set_usize (scrollwin, -1, 200);

   clist = prefs_win.detail_possible =  gtk_clist_new (1);
   gtk_clist_set_selection_mode (GTK_CLIST (clist), GTK_SELECTION_SINGLE);
   gtk_clist_set_column_auto_resize (GTK_CLIST (clist), 0, TRUE);
   gtk_container_add (GTK_CONTAINER (scrollwin), clist);

   for (i = 1; i < detailview_get_titles_num (); i++) {
      gchar *text = _(detailview_get_title (i));
      gtk_clist_append (GTK_CLIST (clist), &text);
      gtk_clist_set_row_data (GTK_CLIST (clist), i - 1, GINT_TO_POINTER (i));
   }

   gtk_signal_connect (GTK_OBJECT (clist),"select_row",
		       GTK_SIGNAL_FUNC (cb_detail_clist_select_row),
		       &prefs_win.detail_possible_selected);
   gtk_signal_connect (GTK_OBJECT (clist),"unselect_row",
		       GTK_SIGNAL_FUNC (cb_detail_clist_unselect_row),
		       &prefs_win.detail_possible_selected);

   prefs_win.detail_possible_selected = -1;

   /* add/delete buttons */
   vbox = gtk_vbox_new (TRUE, 0);
   gtk_table_attach (GTK_TABLE (table), vbox, 1, 2, 1, 2,
		     GTK_FILL, GTK_FILL, 0, 0);

   button1 = gtk_button_new_with_label (_("Add"));
   prefs_win.detail_add = button1;
   gtk_widget_set_usize (button1, 50, -1);
   gtk_box_pack_start (GTK_BOX (vbox), button1, FALSE, FALSE, 2);

   button2 = gtk_button_new_with_label (_("Delete"));
   prefs_win.detail_del = button2;
   gtk_widget_set_usize (button2, 50, -1);
   gtk_box_pack_start (GTK_BOX (vbox), button2, FALSE, FALSE, 2);

   gtk_signal_connect (GTK_OBJECT (button1),"released",
		       GTK_SIGNAL_FUNC (cb_detail_add_button_pressed), clist);
   gtk_signal_connect (GTK_OBJECT (button2),"released",
		       GTK_SIGNAL_FUNC (cb_detail_del_button_pressed), clist);

   hseparator = gtk_hseparator_new ();
   gtk_box_pack_start (GTK_BOX (vbox), hseparator, FALSE, FALSE, 2);

   /* move buttons */
   button1 = gtk_button_new_with_label (_("Up"));
   prefs_win.detail_up = button1;
   gtk_widget_set_usize (button1, 50, -1);
   gtk_box_pack_start (GTK_BOX (vbox), button1, FALSE, FALSE, 2);

   button2 = gtk_button_new_with_label (_("Down"));
   prefs_win.detail_down = button2;
   gtk_widget_set_usize (button2, 50, -1);
   gtk_box_pack_start (GTK_BOX (vbox), button2, FALSE, FALSE, 2);

   /* displayed columns */
   label = gtk_label_new (_("Displayed columns"));
   gtk_table_attach (GTK_TABLE (table), label, 2, 3, 0, 1,
		     0, 0, 0, 0);

   scrollwin = gtk_scrolled_window_new (NULL, NULL);
   gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrollwin),
				   GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
   gtk_container_set_border_width(GTK_CONTAINER(scrollwin), 5);
   gtk_table_attach (GTK_TABLE (table), scrollwin, 2, 3, 1, 2,
		     GTK_FILL | GTK_EXPAND, GTK_EXPAND, 0, 0);
   gtk_widget_set_usize (scrollwin, -1, 200);

   clist = prefs_win.detail_displayed = gtk_clist_new (1);
   gtk_clist_set_selection_mode (GTK_CLIST (clist), GTK_SELECTION_SINGLE);
   gtk_clist_set_column_auto_resize (GTK_CLIST (clist), 0, TRUE);
   gtk_container_add (GTK_CONTAINER (scrollwin), clist);

   gtk_signal_connect (GTK_OBJECT (clist),"select_row",
		       GTK_SIGNAL_FUNC (cb_detail_clist_select_row),
		       &prefs_win.detail_displayed_selected);
   gtk_signal_connect (GTK_OBJECT (clist),"unselect_row",
		       GTK_SIGNAL_FUNC (cb_detail_clist_unselect_row),
		       &prefs_win.detail_displayed_selected);

   prefs_win.detail_displayed_selected = -1;
   detail_set_data (conf.detailview_columns);

   gtk_signal_connect (GTK_OBJECT (button1),"released",
		       GTK_SIGNAL_FUNC (cb_detail_up_button_pressed), clist);
   gtk_signal_connect (GTK_OBJECT (button2),"released",
		       GTK_SIGNAL_FUNC (cb_detail_down_button_pressed), clist);


   /* show/hide column title */
   toggle = create_check_button (_("Show Column Title"),
				 conf.detailview_show_title,
				 cb_get_data_from_toggle,
				 &config_changed->detailview_show_title);
   gtk_box_pack_start (GTK_BOX (main_vbox), toggle, FALSE, FALSE, 0);

   return main_vbox;
}


/*******************************************************************************
 *  prefs_thumbview_mouse_page:
 *     @ Create mouse button action on thumbnail view preference page.
 *
 *  Return : Packed widget (GtkVbox)
 *******************************************************************************/
GtkWidget *
prefs_thumbview_mouse_page ()
{
   GtkWidget *vbox;

   vbox = create_mouse_prefs (thumbview_mouse_items,
			      conf.thumbview_mouse_button,
			      &config_changed->thumbview_mouse_button);

   return vbox;
}


/*******************************************************************************
 *  prefs_dirview_page:
 *     @ Create Directory View preference page.
 *
 *  Return : Packed widget (GtkVbox)
 *******************************************************************************/
GtkWidget *
prefs_dirview_page ()
{
   GtkWidget *main_vbox;
   GtkWidget *frame, *vbox, *hbox;
   GtkWidget *toggle, *label, *option_menu, *spinner;
   GtkAdjustment *adj;

   main_vbox = gtk_vbox_new (FALSE, 0);
   gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 5);


   /**********************************************
    * Shoe/Hide Frame
    **********************************************/
   prefs_create_frame (_("Show/Hide"), frame, vbox, main_vbox);

   /* show/hide toolbar */
   toggle = create_check_button (_("Show toolbar"),
				 conf.dirview_show_toolbar,
                                 cb_get_data_from_toggle,
                                 &config_changed->dirview_show_toolbar);
   gtk_box_pack_start (GTK_BOX (vbox), toggle, FALSE, FALSE, 0);

   /* show/hide dot file */
   toggle = create_check_button (_("Show dot file"),
				 conf.dirview_show_dotfile,
                                 cb_get_data_from_toggle,
                                 &config_changed->dirview_show_dotfile);
   gtk_box_pack_start (GTK_BOX (vbox), toggle, FALSE, FALSE, 0);

   /* show/hide dot file */
   toggle = create_check_button (_("Show \".\" directory whether hide dotfile or not."),
				 conf.dirview_show_current_dir,
                                 cb_get_data_from_toggle,
                                 &config_changed->dirview_show_current_dir);
   gtk_box_pack_start (GTK_BOX (vbox), toggle, FALSE, FALSE, 0);

   /* show/hide dot file */
   toggle = create_check_button (_("Show \"..\" directory whether hide dotfile or not."),
				 conf.dirview_show_parent_dir,
                                 cb_get_data_from_toggle,
                                 &config_changed->dirview_show_parent_dir);
   gtk_box_pack_start (GTK_BOX (vbox), toggle, FALSE, FALSE, 0);

   /* set expander to all */
   toggle = create_check_button (_("Set expander to all directory (Fast)"),
				 conf.dirview_set_expander_all,
                                 cb_get_data_from_toggle,
                                 &config_changed->dirview_set_expander_all);
   gtk_box_pack_start (GTK_BOX (vbox), toggle, FALSE, FALSE, 0);


   /**********************************************
    * Style Frame
    **********************************************/
   prefs_create_frame (_("Style"), frame, vbox, main_vbox);

   /* tree line style */
   hbox = gtk_hbox_new (FALSE, 0);
   gtk_container_set_border_width(GTK_CONTAINER (hbox), 0);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
   label = gtk_label_new (_("Tree line style"));
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 5);
   option_menu = create_option_menu_simple (ctree_line_style_items,
					    conf.dirview_line_style,
					    (gint *) &config_changed->dirview_line_style);
   gtk_widget_set_usize (option_menu, 100, -1);
   gtk_box_pack_start (GTK_BOX (hbox), option_menu, FALSE, FALSE, 5);

   /* tree expander style */
   hbox = gtk_hbox_new (FALSE, 0);
   gtk_container_set_border_width(GTK_CONTAINER (hbox), 0);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
   label = gtk_label_new (_("Tree expander style"));
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 5);
   option_menu = create_option_menu_simple (ctree_expander_style_items,
					    conf.dirview_expander_style,
					    (gint *) &config_changed->dirview_expander_style);
   gtk_widget_set_usize (option_menu, 100, -1);
   gtk_box_pack_start (GTK_BOX (hbox), option_menu, FALSE, FALSE, 5);


   /**********************************************
    * Drag and Drop Frame
    **********************************************/
   prefs_create_frame (_("Drag and Drop"), frame, vbox, main_vbox);

   /* auto scroll */
   toggle = create_check_button (_("Enable auto scroll."),
				 conf.dirview_auto_scroll,
                                 cb_get_data_from_toggle,
                                 &config_changed->dirview_auto_scroll);
   gtk_box_pack_start (GTK_BOX (vbox), toggle, FALSE, FALSE, 0);

   /* auto scroll interval */
   hbox = gtk_hbox_new (FALSE, 5);
   gtk_container_set_border_width (GTK_CONTAINER(hbox), 5);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
   label = gtk_label_new (_("Auto scroll interval"));
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   adj = (GtkAdjustment *) gtk_adjustment_new (conf.dirview_auto_scroll_time,
					       0, 1000,
					       5.0, 10.0, 0.0);
   spinner = gtk_spin_button_new (adj, 0, 0);
   gtk_widget_set_usize(spinner, 70, -1);
   gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (spinner), TRUE);
   gtk_spin_button_set_shadow_type (GTK_SPIN_BUTTON (spinner), GTK_SHADOW_OUT);
   gtk_signal_connect (GTK_OBJECT (adj), "value_changed",
		       GTK_SIGNAL_FUNC (cb_get_data_from_adjustment_by_int),
		       &config_changed->dirview_auto_scroll_time);
   gtk_box_pack_start (GTK_BOX (hbox), spinner, FALSE, FALSE, 0);

   label = gtk_label_new (_("[ms]"));
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);

   /* auto expand */
   toggle = create_check_button (_("Enable auto expand directory tree."),
				 conf.dirview_auto_expand,
                                 cb_get_data_from_toggle,
                                 &config_changed->dirview_auto_expand);
   gtk_box_pack_start (GTK_BOX (vbox), toggle, FALSE, FALSE, 0);
 
   /* auto expand time */
   hbox = gtk_hbox_new (FALSE, 5);
   gtk_container_set_border_width (GTK_CONTAINER(hbox), 5);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
   label = gtk_label_new (_("Lag time for auto expand"));
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   adj = (GtkAdjustment *) gtk_adjustment_new (conf.dirview_auto_expand_time,
					       0, 60000,
					       100.0, 500.0, 0.0);
   spinner = gtk_spin_button_new (adj, 0, 0);
   gtk_widget_set_usize(spinner, 70, -1);
   gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (spinner), TRUE);
   gtk_spin_button_set_shadow_type (GTK_SPIN_BUTTON (spinner), GTK_SHADOW_OUT);
   gtk_signal_connect (GTK_OBJECT (adj), "value_changed",
		       GTK_SIGNAL_FUNC (cb_get_data_from_adjustment_by_int),
		       &config_changed->dirview_auto_expand_time);
   gtk_box_pack_start (GTK_BOX (hbox), spinner, FALSE, FALSE, 0);

   label = gtk_label_new (_("[ms]"));
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);

   return main_vbox;
}


/*******************************************************************************
 *  prefs_preview_page:
 *     @ Create pewview preference page.
 *
 *  Return : Packed widget (GtkVbox)
 *******************************************************************************/
GtkWidget *
prefs_preview_page ()
{
   GtkWidget *main_vbox, *frame, *vbox, *hbox;
   GtkWidget *spinner, *toggle, *label;
   GtkAdjustment *adj;

   main_vbox = gtk_vbox_new (FALSE, 0);
   gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 5);

   /********************************************** 
    * Image Frame
    **********************************************/
   prefs_create_frame (_("Image"), frame, vbox, main_vbox);

   /* Auto resize image to window size */
   toggle = create_check_button (_("Auto resize to frame size when open image"),
				 conf.preview_fit_image_to_frame,
				 cb_fit_image_to_frame, NULL);
   gtk_box_pack_start (GTK_BOX (vbox), toggle, FALSE, FALSE, 0);

   /* Keep Aspect Ratio */
   toggle = create_check_button (_("Keep aspect rario"), conf.preview_keep_aspect,
				 cb_get_data_from_toggle,
				 &config_changed->preview_keep_aspect);
   gtk_box_pack_start (GTK_BOX (vbox), toggle, FALSE, FALSE, 0);

   /* Default Image Scale Spinner */
   hbox = gtk_hbox_new (FALSE, 5);
   gtk_container_set_border_width (GTK_CONTAINER(hbox), 5);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
   label = gtk_label_new (_("Default Image Scale"));
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   adj = (GtkAdjustment *) gtk_adjustment_new (conf.preview_scale,
					       1.0, 10000.0, 1.0, 5.0, 0.0);
   spinner = gtk_spin_button_new (adj, 0, 0);
   gtk_widget_set_usize(spinner, 50, -1);
   prefs_win.preview_scale_spin = spinner;
   gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (spinner), TRUE);
   gtk_spin_button_set_shadow_type (GTK_SPIN_BUTTON (spinner), GTK_SHADOW_OUT);
   gtk_signal_connect (GTK_OBJECT (adj), "value_changed",
		       GTK_SIGNAL_FUNC (cb_get_data_from_adjustment_by_int),
		       &config_changed->preview_scale);
   gtk_box_pack_start (GTK_BOX (hbox), spinner, FALSE, FALSE, 0);
   label = gtk_label_new (_("%"));
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);

   /* Show Scrollbar or not */
   toggle = create_check_button (_("Show scrollbar"),
				 conf.preview_scrollbar,
				 cb_get_data_from_toggle,
				 &config_changed->preview_scrollbar);
   prefs_win.preview_scrollbar_toggle = toggle;
   gtk_box_pack_start (GTK_BOX (vbox), toggle, FALSE, FALSE, 0);

   /* keep original image on memory or not */
   toggle = create_check_button (_("Keep original image on memory"),
				 conf.preview_buffer,
				 cb_get_data_from_toggle,
				 &config_changed->preview_buffer);
   gtk_box_pack_start (GTK_BOX (vbox), toggle, FALSE, FALSE, 0);

   return main_vbox;
}


/*******************************************************************************
 *  prefs_preview_mouse_page:
 *     @ Create mouse button action on preview preference page.
 *
 *  Return : Packed widget (GtkVbox)
 *******************************************************************************/
GtkWidget *
prefs_preview_mouse_page ()
{
   GtkWidget *vbox;

   vbox = create_mouse_prefs (preview_mouse_items,
			      conf.preview_mouse_button,
			      &config_changed->preview_mouse_button);

   return vbox;
}


void
prefs_ui_tumbwin_set_sensitive ()
{
   set_sensitive_thumbwin_state ();
   set_sensitive_detail_view    ();
   set_sensitive_preview        ();
}
