/*
 * GImageView
 * Copyright (C) 2001 Takuro Ashie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * These codes are mostly taken from GQview.
 * GQview code Copyright (C) 2001 John Ellis
 */

#include <stdlib.h>

#include "fileutil.h"
#include "gimv_image.h"
#include "icon.h"
#include "similar.h"

/*
 * These functions are intended to find images with similar color content. For
 * example when an image was saved at different compression levels or dimensions
 * (scaled down/up) the contents are similar, but these files do not match by file
 * size, dimensions, or checksum.
 *
 * These functions create a 32 x 32 array for each color channel (red, green, blue).
 * The array represents the average color of each corresponding part of the
 * image. (imagine the image cut into 1024 rectangles, or a 32 x 32 grid.
 * Each grid is then processed for the average color value, this is what
 * is stored in the array)
 *
 * To compare two images, generate a ImageSimilarityData for each image, then
 * pass them to the compare function. The return value is the percent match
 * of the two images. (for this, simple comparisons are used, basically the return
 * is an average of the corresponding array differences)
 *
 * for image_sim_compare(), the return is 0.0 to 1.0:
 *  1.0 for exact matches (an image is compared to itself)
 *  0.0 for exact opposite images (compare an all black to an all white image)
 * generally only a match of > 0.85 are significant at all, and >.95 is useful to
 * find images that have been re-saved to other formats, dimensions, or compression.
 */

ImageSimilarityData *
image_sim_new (void)
{
   ImageSimilarityData *sd = g_new0 (ImageSimilarityData, 1);

   return sd;
}

void
image_sim_delete (ImageSimilarityData *sd)
{
   g_free (sd);
}

void
image_sim_fill_data (ImageSimilarityData *sd, GimvImage *image)
{
   gint w, h;
   gint rs;
   guchar *pix;
   gint has_alpha;
   gint p_step;

   guchar *p;
   gint i;
   gint j;
   gint x_inc, y_inc;
   gint xs, ys;

   gint x_small = FALSE;	/* if less than 32 w or h, set TRUE */
   gint y_small = FALSE;

   if (!sd || !image) return;

   w = gimv_image_width (image);
   h = gimv_image_height (image);
   rs = gimv_image_rowstride (image);
   pix = gimv_image_get_pixels (image);
   has_alpha = gimv_image_has_alpha (image);

   p_step = has_alpha ? 4 : 3;
   x_inc = w / 32;
   y_inc = h / 32;

   if (x_inc < 1) {
      x_inc = 1;
      x_small = TRUE;
   }
   if (y_inc < 1) {
      y_inc = 1;
      y_small = TRUE;
   }

   j = 0;

   for (ys = 0; ys < 32; ys++) {
      if (y_small) j = (float)h / 32 * ys;

      i = 0;

      for (xs = 0; xs < 32; xs++) {
	 gint x, y;
	 gint r, g, b;

	 if (x_small)
	    i = (float) w / 32 * xs;

	 r = g = b = 0;

	 for (y = j; y < j + y_inc; y++) {
	    p = pix + (y * rs) + (i * p_step);
	    for (x = i; x < i + x_inc; x++) {
	       r += *p; p++;
	       g += *p; p++;
	       b += *p; p++;
	       if (has_alpha) p++;
	    }
	 }
	 r /= x_inc * y_inc;
	 g /= x_inc * y_inc;
	 b /= x_inc * y_inc;

	 sd->avg_r[ys * 32 + xs] = r;
	 sd->avg_g[ys * 32 + xs] = g;
	 sd->avg_b[ys * 32 + xs] = b;

	 i += x_inc;
      }

      j += y_inc;
   }

   sd->filled = TRUE;
}

ImageSimilarityData *
image_sim_new_from_image (GimvImage *image)
{
   ImageSimilarityData *sd;

   sd = image_sim_new ();
   image_sim_fill_data (sd, image);

   return sd;
}


ImageSimilarityData *
image_sim_new_from_thumb (Thumbnail *thumb)
{
   GimvImage *image;
   GdkPixmap *pixmap;
   GdkBitmap *mask;
   ImageSimilarityData *data;

   g_return_val_if_fail (thumb, NULL);

   if (!thumbnail_has_thumbnail (thumb)) return NULL;

   thumbnail_get_thumb (thumb, &pixmap, &mask)     ;
   if (!pixmap) return NULL;;
   image = gimv_image_create_from_drawable (pixmap, 0, 0,
                                            thumb->thumb_width,
                                            thumb->thumb_height);

   if (!image) return NULL;
   data = image_sim_new_from_image (image);

   gimv_image_kill (image);

   return data;
}


gfloat
image_sim_compare (ImageSimilarityData *a, ImageSimilarityData *b)
{
   gfloat sim;
   gint i;

   if (!a || !b || !a->filled || !b->filled) return 0.0;

   sim = 0.0;

   for (i = 0; i < 1024; i++) {
      sim += (float) abs (a->avg_r[i] - b->avg_r[i]) / 255.0;
      sim += (float) abs (a->avg_g[i] - b->avg_g[i]) / 255.0;
      sim += (float) abs (a->avg_b[i] - b->avg_b[i]) / 255.0;
   }

   sim /= (1024.0 * 3.0);

   return 1.0 - sim;
}

/* this uses a cutoff point so that it can abort early when it gets to
 * a point that can simply no longer make the cut-off point.
 */
gfloat
image_sim_compare_fast (ImageSimilarityData *a, ImageSimilarityData *b, gfloat min)
{
   gfloat sim;
   gint i;
   gint j;

   if (!a || !b || !a->filled || !b->filled) return 0.0;

   min = 1.0 - min;
   sim = 0.0;

   for (j = 0; j < 1024; j+= 32) {
      for (i = j; i < j + 32; i++) {
	 sim += (float) abs (a->avg_r[i] - b->avg_r[i]) / 255.0;
	 sim += (float) abs (a->avg_g[i] - b->avg_g[i]) / 255.0;
	 sim += (float) abs (a->avg_b[i] - b->avg_b[i]) / 255.0;
      }
      /* check for abort, if so return 0.0 */
      if (sim / (1024.0 * 3.0) > min) return 0.0;
   }

   sim /= (1024.0 * 3.0);

   return 1.0 - sim;
}



/******************************************************************************
 *
 *   Result Window
 *
 ******************************************************************************/
static gchar *simwin_titles[4] = {
   N_("Name"),
   N_("Accuracy"),
   N_("Size (byte)"),
   N_("Modification Time")
};
static gint simwin_column_num = sizeof (simwin_titles) / sizeof (gchar *);


static gint
cb_window_delete (GtkWidget *widget, GdkEvent *event, ImageSimWin *sw)
{
   g_return_val_if_fail (sw, FALSE);

   return sw->loading;
}


static void
cb_window_destroy (GtkWidget *button, ImageSimWin *sw)
{
   g_return_if_fail (sw);

   g_free (sw);
}


static void
set_pixtext (GtkCTree *ctree, GtkCTreeNode *node, gpointer data)
{
   gboolean thumbnail = GPOINTER_TO_INT (data);
   Thumbnail *thumb;
   GdkPixmap *pixmap;
   GdkBitmap *mask;
   guint8 spacing;
   gboolean is_leaf, expanded;
   gchar *text;

   g_return_if_fail (ctree);
   g_return_if_fail (node);

   thumb = gtk_ctree_node_get_row_data (ctree, node);
   g_return_if_fail (thumb);

   if (thumbnail)
      thumbnail_get_thumb (thumb, &pixmap, &mask);
   else
      thumbnail_get_icon (thumb, &pixmap, &mask);

   gtk_ctree_get_node_info (ctree, node, &text, &spacing,
			    NULL, NULL, NULL, NULL,
                            &is_leaf, &expanded);
   gtk_ctree_set_node_info (ctree, node,
                            text, spacing,
                            pixmap, mask, pixmap, mask,
                            is_leaf, expanded);
}


static void
cb_change_to_thumbnail (GtkWidget *widget, ImageSimWin *sw)
{
   gint i;

   g_return_if_fail (sw);

   for (i = 0; i < simwin_column_num; i++)
      gtk_clist_set_column_auto_resize (GTK_CLIST (sw->ctree), i, TRUE);

   gtk_clist_set_row_height (GTK_CLIST (sw->ctree), sw->thumbnail_size);
   gtk_ctree_post_recursive (GTK_CTREE (sw->ctree), NULL,
                             (GtkCTreeFunc) set_pixtext,
                             GINT_TO_POINTER (TRUE));

   for (i = 0; i < simwin_column_num; i++)
      gtk_clist_set_column_auto_resize (GTK_CLIST (sw->ctree), i, FALSE);

   for (i = 0; i < simwin_column_num; i++)
      gtk_clist_set_column_resizeable (GTK_CLIST (sw->ctree), i, TRUE);
}


static void
cb_change_to_icon (GtkWidget *widget, ImageSimWin *sw)
{
   gint i;

   g_return_if_fail (sw);

   for (i = 0; i < simwin_column_num; i++)
      gtk_clist_set_column_auto_resize (GTK_CLIST (sw->ctree), i, TRUE);

   gtk_clist_set_row_height (GTK_CLIST (sw->ctree), ICON_SIZE);
   gtk_ctree_post_recursive (GTK_CTREE (sw->ctree), NULL,
                             (GtkCTreeFunc) set_pixtext,
                             GINT_TO_POINTER (FALSE));

   for (i = 0; i < simwin_column_num; i++)
      gtk_clist_set_column_auto_resize (GTK_CLIST (sw->ctree), i, FALSE);

   for (i = 0; i < simwin_column_num; i++)
      gtk_clist_set_column_resizeable (GTK_CLIST (sw->ctree), i, TRUE);
}


/* FIXME */
static void
get_filelist (GtkCTree *ctree, GtkCTreeNode *node, GList **filelist)
{
   Thumbnail *thumb;

   g_return_if_fail (ctree);
   g_return_if_fail (node);
   g_return_if_fail (filelist);

   thumb = gtk_ctree_node_get_row_data (ctree, node);
   g_return_if_fail (thumb);

}


static void
cb_open_collection (GtkWidget *button, ImageSimWin *sw)
{
   OpenFiles *files;

   g_return_if_fail (sw);

   files = files_loader_new ();
   gtk_ctree_post_recursive (GTK_CTREE (sw->ctree), NULL,
                             (GtkCTreeFunc) get_filelist,
                             &files->filelist);
   files_loader_delete (files);
}
/* END FIXME */


static void
select_thumb (GtkCTree *ctree, GtkCTreeNode *node, ImageSimWin *sw)
{
   Thumbnail *thumb;

   g_return_if_fail (ctree);
   g_return_if_fail (node);
   g_return_if_fail (sw);
   g_return_if_fail (sw->tv);

   thumb = gtk_ctree_node_get_row_data (ctree, node);
   g_return_if_fail (thumb);

   thumbview_set_selection (thumb, TRUE);
}


static void
cb_select_all (GtkWidget *button, ImageSimWin *sw)
{
   g_return_if_fail (sw);
   if (!sw->tv) return;

   thumbview_set_selection_all (sw->tv, FALSE);
   gtk_ctree_post_recursive (GTK_CTREE (sw->ctree), NULL,
                             (GtkCTreeFunc) select_thumb,
                             sw);
}


static void
cb_window_close (GtkWidget *button, ImageSimWin *sw)
{
   g_return_if_fail (sw);

   if (!sw->loading)
      gtk_widget_destroy (sw->window);
}


void
cb_ctree_select_row (GtkCTree *ctree, GList *node, gint column, ImageSimWin *sw)
{
   Thumbnail *thumb;

   g_return_if_fail (ctree);
   g_return_if_fail (node);
   g_return_if_fail (sw);

   if (!sw->tv) return;

   thumb = gtk_ctree_node_get_row_data (ctree, GTK_CTREE_NODE (node));
   g_return_if_fail (thumb);

   thumbview_set_selection_all (sw->tv, FALSE);
   thumbview_set_selection (thumb, TRUE);
   thumbview_adjust (sw->tv, thumb);
}


ImageSimWin *
image_sim_win_init ()
{
   ImageSimWin *sw;

   sw = g_new0 (ImageSimWin, 1);
   g_return_val_if_fail (sw, NULL);

   sw->window        = NULL;
   sw->ctree         = NULL;
   sw->radio_thumb   = NULL;
   sw->radio_icon    = NULL;
   sw->open_button   = NULL;
   sw->select_button = NULL;
   sw->close_button  = NULL;

   sw->tv            = NULL;

   sw->thumbnail_size = 96;
   sw->loading       = FALSE;

   return sw;
}


ImageSimWin *
image_sim_win_create (gint thumbnail_size)
{
   ImageSimWin *sw = image_sim_win_init ();
   GtkWidget *hbox;
   GtkWidget *scrolledwin, *radio, *button;
   Icon *icon;
   gint i;

   g_return_val_if_fail (sw, NULL);

   sw->thumbnail_size = thumbnail_size;

   /* window */
   sw->window = gtk_dialog_new ();
   gtk_window_set_title (GTK_WINDOW (sw->window), _("Find Duplicates - result")); 
   gtk_window_set_default_size (GTK_WINDOW (sw->window), 500, 400);
   gtk_window_set_position (GTK_WINDOW (sw->window), GTK_WIN_POS_MOUSE);

   gtk_signal_connect (GTK_OBJECT (sw->window), "delete_event",
		       GTK_SIGNAL_FUNC (cb_window_delete), sw);
   gtk_signal_connect (GTK_OBJECT (sw->window), "destroy",
		       GTK_SIGNAL_FUNC (cb_window_destroy), sw);

   /* ctree */
   scrolledwin = gtk_scrolled_window_new (NULL, NULL);
   gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(scrolledwin),
                                   GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
   gtk_container_set_border_width (GTK_CONTAINER (scrolledwin), 5);

   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (sw->window)->vbox),
		       scrolledwin, TRUE, TRUE, 0);

   for (i = 0; i < simwin_column_num; i++)
      simwin_titles[i] = _(simwin_titles[i]);
   sw->ctree = gtk_ctree_new_with_titles (simwin_column_num, 0, simwin_titles);
   gtk_clist_set_column_justification(GTK_CLIST (sw->ctree), 1, GTK_JUSTIFY_CENTER);
   gtk_clist_set_column_justification(GTK_CLIST (sw->ctree), 2, GTK_JUSTIFY_RIGHT);
   gtk_container_add (GTK_CONTAINER (scrolledwin), sw->ctree);

   gtk_signal_connect (GTK_OBJECT (sw->ctree), "tree_select_row",
		       GTK_SIGNAL_FUNC (cb_ctree_select_row), sw);

   /* button */
   hbox = gtk_hbox_new (FALSE, 0);
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (sw->window)->action_area), 
		       hbox, TRUE, TRUE, 0);

   /* radio button */
   radio = gtk_radio_button_new_with_label (NULL, _("Thumbnail"));
   sw->radio_thumb = radio;
   gtk_signal_connect (GTK_OBJECT (radio), "clicked",
		       GTK_SIGNAL_FUNC (cb_change_to_thumbnail), sw);
   gtk_box_pack_start (GTK_BOX (hbox), radio, FALSE, FALSE, 0);

   radio = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON (radio),
                                                        _("Icon"));
   sw->radio_icon  = radio;
   gtk_signal_connect (GTK_OBJECT (radio), "clicked",
		       GTK_SIGNAL_FUNC (cb_change_to_icon), sw);
   gtk_box_pack_start (GTK_BOX (hbox), radio, FALSE, FALSE, 0);

   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radio), TRUE);

   /* Open in New Collection */
   button = gtk_button_new_with_label (_("Open in New Collection"));
   sw->open_button = button;
   gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
   gtk_signal_connect (GTK_OBJECT (button), "clicked",
		       GTK_SIGNAL_FUNC (cb_open_collection), sw);
   GTK_WIDGET_SET_FLAGS(button,GTK_CAN_DEFAULT);
   gtk_widget_show (button);

   /* Select All */
   button = gtk_button_new_with_label (_("Select All"));
   sw->select_button = button;
   gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
   gtk_signal_connect (GTK_OBJECT (button), "clicked",
		       GTK_SIGNAL_FUNC (cb_select_all), sw);
   GTK_WIDGET_SET_FLAGS(button,GTK_CAN_DEFAULT);
   gtk_widget_show (button);

   /* close button */
   button = gtk_button_new_with_label (_("Close"));
   sw->close_button = button;
   gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
   gtk_signal_connect (GTK_OBJECT (button), "clicked",
		       GTK_SIGNAL_FUNC (cb_window_close), sw);
   GTK_WIDGET_SET_FLAGS(button,GTK_CAN_DEFAULT);
   gtk_widget_show (button);

   gtk_widget_grab_focus (button);

   gtk_widget_show_all (sw->window);
   gtk_widget_hide (sw->open_button);

   icon = icon_get ("gimv_icon");
   if (icon)
      gdk_window_set_icon(sw->window->window, NULL, icon->pixmap, icon->mask);

   return sw;
}


void
image_sim_win_set_state (ImageSimWin *sw, gboolean loading)
{
   gint i;

   g_return_if_fail (sw);

   sw->loading = loading;

   {
      GtkWidget *widgets[] = {
         sw->open_button,
         sw->close_button, 
      };

      for (i = 0; i < sizeof (widgets) / sizeof (GtkWidget *); i++)
         gtk_widget_set_sensitive (widgets[i], !sw->loading);
   }

   for (i = 0; i < simwin_column_num; i++)
      gtk_clist_set_column_auto_resize (GTK_CLIST (sw->ctree), i, sw->loading);

   for (i = 0; i < simwin_column_num; i++)
      gtk_clist_set_column_resizeable (GTK_CLIST (sw->ctree), i, !sw->loading);
}


void
image_sim_win_set_relation (ImageSimWin *sw, ThumbView *tv)
{
   g_return_if_fail (sw);
   g_return_if_fail (tv);

   sw->tv = tv;

   gtk_widget_set_sensitive (sw->select_button, TRUE);
}


void
image_sim_win_unset_relation (ImageSimWin *sw)
{
   g_return_if_fail (sw);

   sw->tv = NULL;

   gtk_widget_set_sensitive (sw->select_button, FALSE);
}


static void
cb_node_destroy (gpointer data)
{
   Thumbnail *thumb = data;

   if (thumb)
      thumbnail_unref (thumb);
}


static GtkCTreeNode *
insert_node (ImageSimWin *sw, GtkCTreeNode *parent, Thumbnail *thumb, gfloat similar)
{
   GtkCTreeNode *node;
   GdkPixmap *pixmap;
   GdkBitmap *mask;
   gchar *text[32], accuracy[32];

   g_return_val_if_fail (thumb, NULL);

   if (GTK_TOGGLE_BUTTON (sw->radio_thumb)->active)
      thumbnail_get_thumb (thumb, &pixmap, &mask);
   else
      thumbnail_get_icon (thumb, &pixmap, &mask);

   text[0] = thumbnail_get_image_name (thumb);
   if (similar > 0) {
      g_snprintf (accuracy, 32, "%2.1f%%", similar * 100);
      text[1] = accuracy;
   } else {
      text[1] = NULL;
   }
   text[2] = fileutil_size2str (thumb->info->st.st_size, FALSE);;
   text[3] = fileutil_time2str (thumb->info->st.st_mtime);

   node = gtk_ctree_insert_node (GTK_CTREE (sw->ctree),
                                 parent, NULL, text, 4,
                                 pixmap, mask,
                                 pixmap, mask,
                                 FALSE, FALSE);
   thumbnail_ref (thumb);
   gtk_ctree_node_set_row_data_full (GTK_CTREE (sw->ctree), node,
				     thumb, cb_node_destroy);

   g_free (text[2]);
   g_free (text[3]);

   return node;
}


void
image_sim_win_set_thumb (ImageSimWin *sw,
                         Thumbnail *thumb1,
                         Thumbnail *thumb2,
                         gfloat     similar)
{
   GtkCTreeNode *parent, *node;

   g_return_if_fail (sw);
   g_return_if_fail (thumb1);
   g_return_if_fail (thumb2);

   node = gtk_ctree_find_by_row_data (GTK_CTREE (sw->ctree), NULL, thumb1);
   if (node)
      parent = node;
   else
      parent = insert_node (sw, node, thumb1, -1);

   node = insert_node (sw, parent, thumb2, similar);
   gtk_ctree_expand (GTK_CTREE (sw->ctree), parent);
}
