/*
 * GImageView
 * Copyright (C) 2001 Takuro Ashie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "gimageview.h"

#include "image_window.h"
#include "prefs.h"
#include "slideshow.h"


static gboolean timeout_slideshow     (SlideShow *slideshow);
gint            idle_slideshow_delete (gpointer   data);
static GList   *next_image            (ImageView *iv,
				       gpointer   list_owner,
				       GList     *current,
				       gpointer   data);
static GList   *prev_image            (ImageView *iv,
				       gpointer   list_owner,
				       GList     *current,
				       gpointer   data);
static void     remove_list           (ImageView *iv,
				       gpointer   list_owner,
				       gpointer   data);
static void     set_player            (SlideShow *slideshow);
#if 0
static void     unset_player          (SlideShow *slideshow);
#endif

/******************************************************************************
 *
 *  callback functions
 *
 ******************************************************************************/
void
cb_play_clicked (GtkWidget *button, SlideShow *slideshow)
{
   g_return_if_fail (slideshow);

   slideshow_play (slideshow);
}


void
cb_stop_clicked (GtkWidget *button, SlideShow *slideshow)
{
   g_return_if_fail (slideshow);

   slideshow_stop (slideshow);
}


void
cb_rw_clicked (GtkWidget *button, SlideShow *slideshow)
{
   g_return_if_fail (slideshow);
   g_return_if_fail (slideshow->iw);

   imageview_prev (slideshow->iw->iv);
}


void
cb_fw_clicked (GtkWidget *button, SlideShow *slideshow)
{
   g_return_if_fail (slideshow);
   g_return_if_fail (slideshow->iw);

   imageview_next (slideshow->iw->iv);
}


void
cb_prev_clicked (GtkWidget *button, SlideShow *slideshow)
{
   g_return_if_fail (slideshow);
   g_return_if_fail (slideshow->iw);

   imageview_nth (slideshow->iw->iv, 0);
}


void
cb_next_clicked (GtkWidget *button, SlideShow *slideshow)
{
   gint num;

   g_return_if_fail (slideshow);
   g_return_if_fail (slideshow->iw);

   num = g_list_length (slideshow->filelist);
   imageview_nth (slideshow->iw->iv, num - 1);
}


void
cb_seekbar_pressed (GtkWidget *widget,
		    GdkEventButton *event,
		    SlideShow *slideshow)
{
   ImageWindow *iw;

   g_return_if_fail (slideshow);

   iw = slideshow->iw;
   g_return_if_fail (iw);

   slideshow->seekbar_dragging = TRUE;
}


void
cb_seekbar_released (GtkWidget *widget,
		     GdkEventButton *event,
		     SlideShow *slideshow)
{
   ImageWindow *iw;
   GtkAdjustment *adj;
   gint pos;

   g_return_if_fail (slideshow);

   iw = slideshow->iw;
   g_return_if_fail (iw);

   adj = gtk_range_get_adjustment (GTK_RANGE (iw->player.seekbar));
   pos = adj->value + 0.5;

   imageview_nth (iw->iv, pos);

   slideshow->seekbar_dragging = FALSE;
}



/******************************************************************************
 *
 *  Other private functions
 *
 ******************************************************************************/
/*
 *  timeout_slideshow:
 *     @ Time out function for slideshow.
 *
 *  slideshow : Poiter to SlideShow struct.
 *  Return    : FALSE if end of slideshow or canceled by user.
 */
static gboolean
timeout_slideshow (SlideShow *slideshow)
{
   if ((slideshow->pos && !g_list_next (slideshow->pos) && !slideshow->repeat)
       || !slideshow->running)
   {
      slideshow_stop (slideshow);
      return FALSE;
   }

   imageview_next (slideshow->iw->iv);

   return TRUE;
}


gint
idle_slideshow_delete (gpointer data)
{
   SlideShow *slideshow = data;

   slideshow_delete (slideshow);

   return 0;
}


static GList *
next_image (ImageView *iv,
	    gpointer list_owner,
	    GList *current,
	    gpointer data)
{
   ImageWindow *iw = data;
   GList *next, *node;
   SlideShow *slideshow = list_owner;
   gchar *filename;
   gint num, pos;
   GtkAdjustment *adj;

   g_return_val_if_fail (iv, NULL);
   g_return_val_if_fail (current, NULL);

   node = g_list_find (slideshow->filelist, current->data);
   g_return_val_if_fail (node, NULL);

   next = g_list_next (current);
   if (!next && slideshow->repeat)
      next = g_list_first (current);
   else if (!next)
      next = current;
   g_return_val_if_fail (next, NULL);

   slideshow->pos = next;

   if (next != current) {
      filename = next->data;
      imagewin_change_image_file (iw, filename);

      if (!slideshow->seekbar_dragging) {
	 num = g_list_length (slideshow->filelist);
	 pos = g_list_position (slideshow->filelist, slideshow->pos);
	 adj = gtk_range_get_adjustment (GTK_RANGE (iw->player.seekbar));
	 adj->value          = pos;
	 gtk_adjustment_changed (adj);
      }
   }

   return next;
}


static GList *
prev_image (ImageView *iv,
	    gpointer list_owner,
	    GList *current,
	    gpointer data)
{
   ImageWindow *iw = data;
   GList *prev, *node;
   SlideShow *slideshow = list_owner;
   gchar *filename;
   gint pos;
   GtkAdjustment *adj;

   g_return_val_if_fail (iv, NULL);
   g_return_val_if_fail (current, NULL);

   node = g_list_find (slideshow->filelist, current->data);
   g_return_val_if_fail (node, NULL);

   prev = g_list_previous (current);
   if (!prev && slideshow->repeat)
      prev = g_list_last (current);
   else if (!prev)
      prev = current;
   g_return_val_if_fail (prev, NULL);

   slideshow->pos = prev;

   if (prev != current) {
      filename = prev->data;
      imagewin_change_image_file (iw, filename);

      if (!slideshow->seekbar_dragging) {
	 pos = g_list_position (slideshow->filelist, slideshow->pos);
	 adj = gtk_range_get_adjustment (GTK_RANGE (iw->player.seekbar));
	 adj->value = pos;
	 gtk_adjustment_changed (adj);
      }
   }

   return prev;
}


static GList *
nth_image (ImageView *iv,
	   gpointer list_owner,
	   GList *list,
	   guint nth,
	   gpointer data)
{
   ImageWindow *iw = data;
   GList *node;
   SlideShow *slideshow = list_owner;
   gchar *filename;
   gint pos;
   GtkAdjustment *adj;

   g_return_val_if_fail (iv, NULL);

   node = g_list_nth (slideshow->filelist, nth);
   g_return_val_if_fail (node, NULL);

   slideshow->pos = node;

   filename = node->data;
   imagewin_change_image_file (iw, filename);

   pos = g_list_position (slideshow->filelist, slideshow->pos);
   adj = gtk_range_get_adjustment (GTK_RANGE (iw->player.seekbar));
   adj->value = pos;
   gtk_adjustment_changed (adj);

   return node;
}


static void
remove_list (ImageView *iv, gpointer list_owner, gpointer data)
{
   SlideShow *slideshow = list_owner;

   g_return_if_fail (iv);
   g_return_if_fail (slideshow);

   slideshow->destroy = TRUE;

   if (slideshow->timer_id)
      slideshow_stop (slideshow);

   /* free slide show data safely */
   gtk_idle_add (idle_slideshow_delete, slideshow);
}


static void
set_player (SlideShow *slideshow)
{
   ImageWindow *iw;
   gint num, pos;
   GtkAdjustment *adj;

   g_return_if_fail (slideshow);
   g_return_if_fail (slideshow->iw);

   iw = slideshow->iw;

   num = g_list_length (slideshow->filelist);
   pos = g_list_position (slideshow->filelist, slideshow->pos);

   adj = gtk_range_get_adjustment (GTK_RANGE (iw->player.seekbar));

   adj->lower          = 0.0;
   adj->upper          = num;
   adj->value          = pos;
   adj->step_increment = 1.0;
   adj->page_increment = 1.0;
   adj->page_size      = 1.0;
   gtk_adjustment_changed (adj);

   gtk_signal_connect (GTK_OBJECT(iw->player.play),"clicked",
		       GTK_SIGNAL_FUNC(cb_play_clicked), slideshow);
   gtk_signal_connect (GTK_OBJECT(iw->player.stop),"clicked",
		       GTK_SIGNAL_FUNC(cb_stop_clicked), slideshow);
   gtk_signal_connect (GTK_OBJECT(iw->player.rw),"clicked",
		       GTK_SIGNAL_FUNC(cb_rw_clicked), slideshow);
   gtk_signal_connect (GTK_OBJECT(iw->player.fw),"clicked",
		       GTK_SIGNAL_FUNC(cb_fw_clicked), slideshow);
   gtk_signal_connect (GTK_OBJECT(iw->player.prev),"clicked",
		       GTK_SIGNAL_FUNC(cb_prev_clicked), slideshow);
   gtk_signal_connect (GTK_OBJECT(iw->player.next),"clicked",
		       GTK_SIGNAL_FUNC(cb_next_clicked), slideshow);
   gtk_signal_connect (GTK_OBJECT(iw->player.seekbar),"button_press_event",
		       GTK_SIGNAL_FUNC(cb_seekbar_pressed), slideshow);
   gtk_signal_connect (GTK_OBJECT(iw->player.seekbar),"button_release_event",
		       GTK_SIGNAL_FUNC(cb_seekbar_released), slideshow);

   gtk_widget_set_sensitive (iw->player.play, TRUE);
   gtk_widget_set_sensitive (iw->player.rw, TRUE);
   gtk_widget_set_sensitive (iw->player.fw, TRUE);
   gtk_widget_set_sensitive (iw->player.prev, TRUE);
   gtk_widget_set_sensitive (iw->player.next, TRUE);
   gtk_widget_set_sensitive (iw->player.seekbar, TRUE);
}


#if 0
static void
unset_player (SlideShow *slideshow)
{
   ImageWindow *iw;

   g_return_if_fail (slideshow);
   g_return_if_fail (slideshow->iw);

   iw = slideshow->iw;

   imagewin_player_set_sensitive_all (iw, FALSE);
}
#endif



/******************************************************************************
 *
 *  Public functions
 *
 ******************************************************************************/
SlideShow *
slideshow_new ()
{
   SlideShow *slideshow;

   slideshow = g_new0 (SlideShow, 1);

   slideshow->iw               = NULL;
   slideshow->filelist         = NULL;
   slideshow->pos              = NULL;
   slideshow->repeat           = conf.slideshow_repeat;
   slideshow->running          = FALSE;
   slideshow->seekbar_dragging = FALSE;
   slideshow->destroy          = FALSE;

   return slideshow;
}


void
slideshow_delete (SlideShow *slideshow)
{
   ImageWindow *iw;
   ImageView *iv;

   g_return_if_fail (slideshow);

   iw = slideshow->iw;
   iv = iw->iv;

   iw->slideshow = NULL;
   g_list_foreach (slideshow->filelist, (GFunc) g_free, NULL);
   g_list_free (g_list_first (slideshow->filelist));

   g_free (slideshow);
}


SlideShow *
slideshow_new_with_filelist (GList *filelist)
{
   SlideShow *slideshow;

   g_return_val_if_fail (filelist, NULL);

   slideshow           = slideshow_new ();
   slideshow->filelist = g_list_first (filelist);
   slideshow->pos      = g_list_first (filelist);
   slideshow->interval = conf.slideshow_interval * 1000;
   slideshow->timer_id = 0;

   return slideshow;
}


ImageWindow *
slideshow_open_window (SlideShow *slideshow)
{
   ImageWindow *iw;
   ImageView *iv;
   gchar *filename;
   GdkColor color;
   GdkColormap *colormap;

   g_return_val_if_fail (slideshow, NULL);
   g_return_val_if_fail (slideshow->filelist, NULL);

   if (!slideshow->pos)
      slideshow->pos = g_list_first (slideshow->filelist);
   g_return_val_if_fail (slideshow->pos, NULL);

   filename = slideshow->pos->data;
   slideshow->iw = imagewin_initialize (filename);
   g_return_val_if_fail (slideshow->iw, NULL);

   iw = slideshow->iw;
   iv = iw->iv;

   /* override some parameters */
   iv->fit_to_frame     = conf.slideshow_fit_to_win;
   iv->x_scale          = conf.slideshow_img_scale;
   iv->y_scale          = conf.slideshow_img_scale;
   iv->keep_aspect      = conf.slideshow_keep_aspect;

   iw->show_menubar     = conf.slideshow_menubar;
   iw->show_toolbar     = conf.slideshow_toolbar;
   iw->show_player      = conf.slideshow_player;
   iw->show_statusbar   = conf.slideshow_statusbar;
   iw->fullscreen       = conf.slideshow_fullscreen;
   iw->hide_frame       = TRUE;
   iw->slideshow        = slideshow;

   imagewin_create_window (slideshow->iw);

   /* set bg color */
   if (conf.slideshow_set_bg) {
      /*
      gint i;
      for (i = 0; i < 3; i++) {
	 if (conf.slideshow_bg_color[i] < 0)
	    conf.slideshow_bg_color[i] = 0;
	 if (conf.slideshow_bg_color[i] > 256)
	    conf.slideshow_bg_color[i] = 255;
      }
      */
      color.pixel = 0;
      color.red   = conf.slideshow_bg_color[0];
      color.green = conf.slideshow_bg_color[1];
      color.blue  = conf.slideshow_bg_color[2];
      colormap = gdk_window_get_colormap(iv->draw_area->window);
      gdk_colormap_alloc_color (colormap, &color, FALSE, TRUE);
      imageview_set_bg_color (iv, &color);
   }

   imageview_set_list (iv,
		       slideshow->filelist,
		       slideshow->pos,
		       (gpointer) slideshow,
		       next_image, iw,
		       prev_image, iw,
		       nth_image,  iw,
		       remove_list, NULL);
   set_player (slideshow);

   return iw;
}


/*
 *  slideshow_play:
 *     @ Execute slide show;
 *
 *  slideshow : Poiter to SlideShow struct.
 */
void
slideshow_play (SlideShow *slideshow)
{
   ImageWindow *iw = NULL;

   g_return_if_fail (slideshow);

   if (!slideshow->iw)
      iw = slideshow_open_window (slideshow);
   else
      iw = slideshow->iw;

   g_return_if_fail (iw);

   if (!slideshow->pos)
      slideshow->pos = g_list_first (slideshow->filelist);
   g_return_if_fail (slideshow->pos);

   slideshow->running = TRUE;
   slideshow->timer_id
      = gtk_timeout_add (slideshow->interval,
			 (GtkFunction) timeout_slideshow,
			 (gpointer) slideshow);

   gtk_widget_set_sensitive (iw->player.play, FALSE);
   gtk_widget_set_sensitive (iw->player.stop, TRUE);
}


void
slideshow_stop (SlideShow *slideshow)
{
   ImageWindow *iw;

   g_return_if_fail (slideshow);

   iw = slideshow->iw;

   slideshow->running = FALSE;

   if (slideshow->timer_id)
      gtk_timeout_remove (slideshow->timer_id);

   if (!slideshow->destroy) {
      gtk_widget_set_sensitive (iw->player.play, TRUE);
      gtk_widget_set_sensitive (iw->player.stop, FALSE);
   }
}


void
slideshow_set_interval (SlideShow *slideshow, guint interval)
{
   slideshow->interval = interval;
   if (slideshow->running) {
      slideshow_stop (slideshow);
      slideshow_play (slideshow);
   }
}
