/*
 * GImageView
 * Copyright (C) 2001 Takuro Ashie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>

#include "gimageview.h"
#include "gimv_image.h"
#include "fileload.h"
#include "fileutil.h"
#include "fr-archive.h"
#include "gfileutil.h"
#include "icon.h"
#include "image_info.h"
#include "prefs.h"
#include "thumbnail.h"
#include "thumbnail_support.h"
#include "thumbnail_view.h"


/* private functions */
static void       store_thumbnail                (GimvImage      *image,
						  gint            thumbsize,
						  GdkPixmap     **pixmap,
						  GdkBitmap     **mask,
						  gint           *width,
						  gint           *height);
static void       create_thumbnail               (Thumbnail      *thumb,
						  gint            thumbsize,
						  gchar          *cache_type,
						  gboolean        save_cache);
static void       load_thumbnail_cache           (Thumbnail      *thumb,
						  gint            thumbsize);
static void       thumbnail_remove_mode_data     (Thumbnail *thumb);


static GHashTable *cache_type_labels;
static gchar *config_cache_read_string = NULL;
static GList *cache_read_list = NULL;


/******************************************************************************
 *
 *   Private Functions.
 *
 ******************************************************************************/
/* defined and initialized in thumbnail_view.c */
extern ThumbDispMode thumb_view_modes[];

static void
thumbnail_remove_mode_data (Thumbnail *thumb)
{
   GList *list;

   g_return_if_fail (thumb);

   list = thumbview_get_disp_mode_list ();
   while (list) {
      ThumbDispMode *view = list->data;
      if (view && view->thumbnail_data_remove_func)
	 view->thumbnail_data_remove_func (thumb);
      list = g_list_next (list);
   }
}


static void
store_thumbnail (GimvImage *image,   gint thumbsize,
		 GdkPixmap **pixmap, GdkBitmap **mask,
		 gint *width_ret, gint *height_ret)
{
   gfloat scale;
   gint width = 0, height = 0;

   if (!image || !thumbsize) return;
   g_return_if_fail (pixmap && mask);

   if (width_ret)  *width_ret = -1;
   if (height_ret) *height_ret = -1;

   if (gimv_image_width (image) >= gimv_image_height (image)) {
      scale = (gfloat) thumbsize / (gfloat) gimv_image_width (image);
      width = thumbsize;
      height = gimv_image_height (image) * scale;
   } else {
      scale = (gfloat) thumbsize / (gfloat) gimv_image_height (image);
      width = gimv_image_width (image) * scale;
      height = thumbsize;
   }

   if (width > 0 && height > 0) {
      gimv_image_scale_get_pixmap (image, width, height, pixmap, mask);
      if (width_ret)  *width_ret = width;
      if (height_ret) *height_ret = height;
   }
}


static void
thumbnail_apply_cache_read_prefs ()
{
   gchar **labels;
   gint i;

   if (cache_read_list) {
      g_list_foreach (cache_read_list, (GFunc)g_free, NULL);
      g_list_free (cache_read_list);
      cache_read_list = NULL;
   }

   config_cache_read_string = conf.cache_read_list;
   if (!conf.cache_read_list) return;

   labels = g_strsplit (conf.cache_read_list, ",", -1);

   for (i = 0; labels[i]; i++) {
      cache_read_list = g_list_append (cache_read_list, g_strdup (labels[i]));
   }

   g_strfreev (labels);
}


/*
 *  create_thumbnail:
 *     @ Render thumbnail from original image or thumbnail cache,
 *       and create GtkPixmap Widget.
 *     @ If save_cache equal TRUE and specified thumbnail type,
 *       call save_thumbnail_cache () to save thumbnail to disk.
 *
 *  thumb     : Pointer to Thumbnail struct.
 *  thumbsize : Size of thumbnail. Positive value means thumbnail size is
 *              determined by longer axis, and negative value means determined
 *              by longer axis.
 *  type      : Thumbnail cache type.
 *  Return    : Pointer to GtkPixmap widget(thumbnail).
 */
static void
create_thumbnail (Thumbnail *thumb, gint thumbsize,
		  gchar *cache_type, gboolean save_cache)
{
   ImageInfo *info;
   GimvImage *image;
   gboolean saved = FALSE;
   ThumbCacheSaveInfo *save_info;
   gchar *filename;


   g_return_if_fail (thumb);
   g_return_if_fail (thumb->info);

   info = thumb->info;

   /* get filename */
   filename = image_info_get_path_with_archive (info);
   if (!filename)
      filename = g_strdup (info->filename);

   /* load cache if cache type is specified */
   if (cache_type && !save_cache) {
      image = thumbsupport_load_thumbnail_cache (filename, cache_type);
      saved = TRUE;

   /* load original image */
   } else {
      gchar *tempfile;
      if (image_info_is_in_archive (info)) {
	 tempfile = image_info_extract_archive (info);
	 if (!tempfile) goto ERROR;
	 if (info->archive->process->error != FR_PROC_ERROR_NONE) 
	    goto ERROR;
	 image = gimv_image_load_file (tempfile);
	 g_free (tempfile);
      } else {
	 image = gimv_image_load_file (filename);
      }
   }

   if (image) {
      /* write cache to disk */
      if (save_cache && cache_type
	  && (!image_info_is_in_archive (info)
	      || !strcmp (cache_type, "GImageView")))
      {
	 image_info_set_data_from_image (info, image);
	 save_info = thumbsupport_save_info_new (info, NULL);
	 saved = thumbsupport_save_thumbnail_cache (filename, cache_type,
						    image, save_info);
	 thumbsupport_save_info_delete (save_info);
      }

      /* resize to display thumbnail size */
      store_thumbnail (image, thumbsize,
		       &thumb->thumbnail,
		       &thumb->thumbnail_mask,
		       &thumb->thumb_width,
		       &thumb->thumb_height);
      store_thumbnail (image, ICON_SIZE,
		       &thumb->icon,
		       &thumb->icon_mask,
		       NULL,
		       NULL);

      gimv_image_kill  (image);
   }

   if (thumb->thumbnail && saved && cache_type) {
      gchar *label;

      if (!cache_type_labels)
	 cache_type_labels = g_hash_table_new (g_str_hash, g_str_equal);
      label = g_hash_table_lookup (cache_type_labels, cache_type);
      if (!label) {
	 label = g_strdup (cache_type);
	 g_hash_table_insert (cache_type_labels, label, label);
      }
      thumb->cache_type = label;
   }

 ERROR:
   g_free (filename);
}


/*
 *  load_thumbnail_cache:
 *     @ Create cache from thumbnail disk cache. 
 *
 *  thumb     : Pointer to Thumbnail struct.
 *  thumbsize : Size of thumbnail.
 *  Return    : Pointer to GtkPixmap widget(thumbnail).
 */
static void
load_thumbnail_cache (Thumbnail *thumb, gint thumbsize)
{
   GList *node;

   g_return_if_fail (thumb);

   if (!cache_read_list || config_cache_read_string != conf.cache_read_list)
      thumbnail_apply_cache_read_prefs ();

   node = cache_read_list;
   while (node) {
      gchar *cache_type = node->data;
      create_thumbnail (thumb, thumbsize, cache_type, FALSE);
      if (thumb->thumbnail) return;
      node = g_list_next (node);
   }
}


/******************************************************************************
 *
 *   Public Functions.
 *
 ******************************************************************************/
Thumbnail *
thumbnail_new (const gchar *filename)
{
   Thumbnail *thumb;

   thumb = g_new0 (Thumbnail, 1);
   g_return_val_if_fail (thumb, NULL);

   thumb->info = image_info_get (filename);
   thumb->thumb_view = NULL;
   thumb->selected = FALSE;
   thumb->sync = FALSE;
   thumb->mode_data = g_hash_table_new (g_str_hash, g_str_equal);
   thumb->ref_count = 0;

   thumb->image          = NULL;
   thumb->thumbnail      = NULL;
   thumb->thumbnail_mask = NULL;
   thumb->icon           = NULL;
   thumb->icon_mask      = NULL;
   thumb->thumb_width    = -1;
   thumb->thumb_height   = -1;
   thumb->cache_type     = NULL;

   thumb = thumbnail_ref (thumb);

   return thumb;
}


Thumbnail *
thumbnail_new_with_image_info (ImageInfo *info)
{
   Thumbnail *thumb;

   thumb = g_new0 (Thumbnail, 1);
   g_return_val_if_fail (thumb, NULL);
   g_return_val_if_fail (info,  NULL);

   thumb->info = image_info_ref (info);
   thumb->thumb_view = NULL;
   thumb->selected = FALSE;
   thumb->sync = FALSE;
   thumb->mode_data = g_hash_table_new (g_str_hash, g_str_equal);
   thumb->ref_count = 0;

   thumb->image          = NULL;
   thumb->thumbnail      = NULL;
   thumb->thumbnail_mask = NULL;
   thumb->icon           = NULL;
   thumb->icon_mask      = NULL;
   thumb->thumb_width    = -1;
   thumb->thumb_height   = -1;
   thumb->cache_type     = NULL;

   thumb = thumbnail_ref (thumb);

   return thumb;
}


ThumbView *
thumbnail_get_parent_thumbview (Thumbnail *thumb)
{
   g_return_val_if_fail (thumb, NULL);

   return thumb->thumb_view;
}


void
thumbnail_set_parent_thumbview (Thumbnail *thumb, ThumbView *tv)
{
   g_return_if_fail (thumb);

   thumb->thumb_view = tv;
}


void
thumbnail_finalize (Thumbnail *thumb)
{
   g_return_if_fail (thumb);

   /* free thumbnail */
   if (thumb->image) {
      gimv_image_kill (thumb->image);
      thumb->image = NULL;
   }

   if (thumb->thumbnail) {
      gdk_pixmap_unref (thumb->thumbnail);
      thumb->thumbnail = NULL;
   }
   if (thumb->thumbnail_mask) {
      gdk_bitmap_unref (thumb->thumbnail_mask);
      thumb->thumbnail_mask = NULL;
   }
   if (thumb->icon) {
      gdk_pixmap_unref (thumb->icon);
      thumb->icon = NULL;
   }
   if (thumb->icon_mask) {
      gdk_bitmap_unref (thumb->icon_mask);
      thumb->icon_mask = NULL;
   }

   image_info_unref (thumb->info);
   thumb->info = NULL;
   thumbnail_remove_mode_data (thumb);
   g_hash_table_destroy (thumb->mode_data);
   g_free (thumb);
}


Thumbnail *
thumbnail_ref (Thumbnail *thumb)
{
   g_return_val_if_fail (thumb, NULL);

   thumb->ref_count++;

   return thumb;
}


void
thumbnail_unref (Thumbnail *thumb)
{
   g_return_if_fail (thumb);

   thumb->ref_count--;
   if (thumb->ref_count < 1)
      thumbnail_finalize (thumb);
}


gchar *
thumbnail_get_image_name (Thumbnail *thumb)
{
   g_return_val_if_fail (thumb, NULL);
   g_return_val_if_fail (thumb->info, NULL);

   return thumb->info->filename;
}


/*
 *  thumbnail_create:
 *     @ Create thumbnail. If disk cache is not exist, create it automaticary.
 *
 *  thumb     : Pointer to Thumbnail struct.
 *  type      : Thumbnail load type
 *              (from cache or original image (or other method...?)).
 *  Return    : Pointer to GtkPixmap widget(thumbnail).
 */
gboolean
thumbnail_create (Thumbnail    *thumb,
		  gint          thumb_size,
		  ThumbLoadType type)
{
   /* free old thumbnail */
   if (thumb->thumbnail) {
      gdk_pixmap_unref (thumb->thumbnail);
      thumb->thumbnail = NULL; thumb->thumbnail_mask = NULL;
   }
   if (thumb->icon) {
      gdk_pixmap_unref (thumb->icon);
      thumb->icon = NULL; thumb->icon_mask = NULL;
   }

   /* if file is directory */
   if (isdir (thumbnail_get_image_name (thumb)))
   {
      Icon *icon;

      icon = icon_get ("folder48");
      thumb->thumbnail      = gdk_pixmap_ref (icon->pixmap);
      thumb->thumbnail_mask = gdk_bitmap_ref (icon->mask);

      icon = icon_get ("folder");
      thumb->icon      = gdk_pixmap_ref (icon->pixmap);
      thumb->icon_mask = gdk_bitmap_ref (icon->mask);
   }

   /* if file is archive */
   if (!thumb->thumbnail && conf.detect_filetype_by_ext
       && fr_archive_utils_get_file_name_ext (thumbnail_get_image_name (thumb)))
   {
      Icon *icon;

      icon = icon_get ("archive");
      thumb->thumbnail      = gdk_pixmap_ref (icon->pixmap);
      thumb->thumbnail_mask = gdk_bitmap_ref (icon->mask);

      icon = icon_get ("small_archive");
      thumb->icon      = gdk_pixmap_ref (icon->pixmap);
      thumb->icon_mask = gdk_bitmap_ref (icon->mask);
   }

   /* load chache  */
   if (!thumb->thumbnail && type != CREATE_THUMB) {
      load_thumbnail_cache (thumb, thumb_size);
   }

   /* create cache if not exist */
   if (!thumb->thumbnail
       && conf.cache_write_type && conf.cache_write_type[0] != '\0')
   {
      create_thumbnail (thumb, thumb_size,
			conf.cache_write_type, TRUE);
   }

   /* create thumbnail from original image */
   if (!thumb->thumbnail) {
      create_thumbnail (thumb, thumb_size, NULL, FALSE);
   }

   if (!thumb->thumbnail)
      return FALSE;
   else
      return TRUE;
}


void
thumbnail_get_thumb (Thumbnail *thumb, GdkPixmap **pixmap, GdkBitmap **mask)
{
   g_return_if_fail (thumb);
   g_return_if_fail (pixmap && mask);

   *pixmap = thumb->thumbnail;
   *mask   = thumb->thumbnail_mask;
}


GtkWidget *
thumbnail_get_thumb_by_widget (Thumbnail *thumb)
{
   g_return_val_if_fail (thumb, NULL);
   if (!thumb->thumbnail) return NULL;

   return gtk_pixmap_new (thumb->thumbnail, thumb->thumbnail_mask);
}


void
thumbnail_get_icon (Thumbnail *thumb, GdkPixmap **pixmap, GdkBitmap **mask)
{
   g_return_if_fail (thumb);
   g_return_if_fail (pixmap && mask);

   *pixmap = thumb->icon;
   *mask   = thumb->icon_mask;
}


GtkWidget *
thumbnail_get_icon_by_widget (Thumbnail *thumb)
{
   g_return_val_if_fail (thumb, NULL);
   if (!thumb->icon) return NULL;

   return gtk_pixmap_new (thumb->icon, thumb->icon_mask);
}


gchar *
thumbnail_get_cache_type (Thumbnail *thumb)
{
   g_return_val_if_fail (thumb, NULL);

   return thumb->cache_type;
}


/*
 *  thumbnail_extract_archive:
 *     @ extract an image from archive.
 *
 *  thumb   :
 *  archive :
 *  Return  : Temporary file name.
 */
gchar *
thumbnail_extract_archive (Thumbnail *thumb)
{
   g_return_val_if_fail (thumb, NULL);
   g_return_val_if_fail (thumb->info, NULL);

   return image_info_extract_archive (thumb->info);
}


gboolean
thumbnail_rename_image (Thumbnail *thumb, const gchar *filename)
{
   g_return_val_if_fail (thumb, FALSE);
   g_return_val_if_fail (thumb->info, FALSE);
   g_return_val_if_fail (filename, FALSE);

   return image_info_rename_image (thumb->info, filename);
}


gboolean
thumbnail_has_thumbnail (Thumbnail *thumb)
{
   g_return_val_if_fail (thumb, FALSE);

   if (isdir (thumbnail_get_image_name (thumb)))
      return FALSE;

   if (fr_archive_utils_get_file_name_ext (thumbnail_get_image_name (thumb)))
      return FALSE;

   if (thumb->thumbnail)
      return TRUE;
   else
      return FALSE;
}


gboolean
thumbnail_is_dir (Thumbnail *thumb)
{
   g_return_val_if_fail (thumb, FALSE);

   if (isdir (thumbnail_get_image_name (thumb)))
      return TRUE;

   return FALSE;
}


gboolean
thumbnail_is_archive (Thumbnail *thumb)
{
   g_return_val_if_fail (thumb, FALSE);

   if (fr_archive_utils_get_file_name_ext (thumbnail_get_image_name (thumb)))
      return TRUE;

   return FALSE;
}


gboolean
thumbnail_is_in_archive (Thumbnail *thumb)
{
   g_return_val_if_fail (thumb, FALSE);

   return image_info_is_in_archive (thumb->info);
}


gchar *
thumbnail_find_thumbcache (const gchar *filename, gchar **type)
{
   gchar *thumb_file = NULL;
   gchar *cache_type;
   GList *node;

   node = cache_read_list;
   while (node) {
      cache_type = node->data;
      thumb_file = thumbsupport_get_thumb_cache_path (filename, cache_type);
      if (file_exists (thumb_file)) {
	 if (type)
	    *type = cache_type;
	 return thumb_file;
      }
      g_free (thumb_file);
      thumb_file = NULL;

      node = g_list_next (node);
   }

   if (type)
      *type = NULL;

   return NULL;
}
